# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from stestr.tests import base
from stestr import utils


class TestUtils(base.TestCase):
    def test_cleanup_test_name_defaults(self):
        test_id_no_attrs = "test.TestThing.test_thing"
        test_id_with_attrs = "test.TestThing.test_thing[attr1,attr2,att3]"
        test_id_with_scenario = "test.TestThing.test_thing(mysql)"
        test_id_with_attrs_and_scenario = "test.TestThing.test_thing[attr]" "(mysql)"
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(test_id_with_attrs)
        self.assertEqual(test_id_no_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(test_id_with_scenario)
        self.assertEqual(test_id_with_scenario, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario
        )
        self.assertEqual(test_id_with_scenario, result_with_attr_and_scenario)

    def test_cleanup_test_name_leave_attrs(self):
        test_id_no_attrs = "test.TestThing.test_thing"
        test_id_with_attrs = "test.TestThing.test_thing[attr1,attr2,att3]"
        test_id_with_scenario = "test.TestThing.test_thing(mysql)"
        test_id_with_attrs_and_scenario = "test.TestThing.test_thing[attr]" "(mysql)"
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs, strip_tags=False)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(
            test_id_with_attrs, strip_tags=False
        )
        self.assertEqual(test_id_with_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(
            test_id_with_scenario, strip_tags=False
        )
        self.assertEqual(test_id_with_scenario, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_tags=False
        )
        self.assertEqual(test_id_with_attrs_and_scenario, result_with_attr_and_scenario)

    def test_cleanup_test_name_strip_scenario_and_attrs(self):
        test_id_no_attrs = "test.TestThing.test_thing"
        test_id_with_attrs = "test.TestThing.test_thing[attr1,attr2,att3]"
        test_id_with_scenario = "test.TestThing.test_thing(mysql)"
        test_id_with_attrs_and_scenario = "test.TestThing.test_thing[attr]" "(mysql)"
        result_no_attrs = utils.cleanup_test_name(
            test_id_no_attrs, strip_scenarios=True
        )
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(
            test_id_with_attrs, strip_scenarios=True
        )
        self.assertEqual(test_id_no_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(
            test_id_with_scenario, strip_scenarios=True
        )
        self.assertEqual(test_id_no_attrs, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_scenarios=True
        )
        self.assertEqual(test_id_no_attrs, result_with_attr_and_scenario)

    def test_cleanup_test_name_strip_scenario(self):
        test_id_no_attrs = "test.TestThing.test_thing"
        test_id_with_attrs = "test.TestThing.test_thing[attr1,attr2,att3]"
        test_id_with_scenario = "test.TestThing.test_thing(mysql)"
        test_id_with_attrs_and_scenario = "test.TestThing.test_thing[attr]" "(mysql)"
        result_no_attrs = utils.cleanup_test_name(
            test_id_no_attrs, strip_scenarios=True, strip_tags=False
        )
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(
            test_id_with_attrs, strip_scenarios=True, strip_tags=False
        )
        self.assertEqual(test_id_with_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(
            test_id_with_scenario, strip_scenarios=True, strip_tags=False
        )
        self.assertEqual(test_id_no_attrs, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_scenarios=True, strip_tags=False
        )
        self.assertEqual(
            "test.TestThing.test_thing[attr]", result_with_attr_and_scenario
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from stestr.commands import slowest
from stestr.tests import base


class TestSlowest(base.TestCase):
    def test_format_times(self):
        times = [("test_id_a", 12.34), ("test_id_b", 1.34)]
        res = slowest.format_times(times)
        self.assertEqual([("test_id_a", "12.340"), ("test_id_b", " 1.340")], res)

    def test_format_times_with_zero(self):
        times = [("test_id_a", 0), ("test_id_b", 1.34)]
        res = slowest.format_times(times)
        self.assertEqual([("test_id_a", "0.000"), ("test_id_b", "1.340")], res)

    def test_format_times_all_zero(self):
        times = [("test_id_a", 0), ("test_id_b", 0.00)]
        res = slowest.format_times(times)
        self.assertEqual([("test_id_a", "0.000"), ("test_id_b", "0.000")], res)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import io
import os
import shutil
import subprocess
import tempfile

import subunit as subunit_lib
import testtools
import yaml

from stestr.tests import base


class TestReturnCodes(base.TestCase):
    def setUp(self):
        super().setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix="stestr-unit")
        self.addCleanup(shutil.rmtree, self.directory)
        self.test_dir = os.path.join(self.directory, "tests")
        os.mkdir(self.test_dir)
        # Setup Test files
        self.testr_conf_file = os.path.join(self.directory, ".stestr.conf")
        self.setup_cfg_file = os.path.join(self.directory, "setup.cfg")
        self.passing_file = os.path.join(self.test_dir, "test_passing.py")
        self.failing_file = os.path.join(self.test_dir, "test_failing.py")
        self.init_file = os.path.join(self.test_dir, "__init__.py")
        self.setup_py = os.path.join(self.directory, "setup.py")
        shutil.copy("stestr/tests/files/testr-conf", self.testr_conf_file)
        shutil.copy("stestr/tests/files/passing-tests", self.passing_file)
        shutil.copy("stestr/tests/files/failing-tests", self.failing_file)
        shutil.copy("stestr/tests/files/setup.py", self.setup_py)
        shutil.copy("stestr/tests/files/setup.cfg", self.setup_cfg_file)
        shutil.copy("stestr/tests/files/__init__.py", self.init_file)

        self.stdout = io.StringIO()
        self.stderr = io.StringIO()
        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, os.path.abspath(os.curdir))
        os.chdir(self.directory)
        subprocess.call("stestr init", shell=True)

    def _get_cmd_stdout(self, cmd):
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        out = p.communicate()
        self.assertEqual(0, p.returncode)
        return out

    def assertRunExit(self, cmd, expected, subunit=False, stdin=None):
        if stdin:
            p = subprocess.Popen(
                "%s" % cmd,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, err = p.communicate(stdin)
        else:
            p = subprocess.Popen(
                "%s" % cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = p.communicate()

        if not subunit:
            self.assertEqual(
                p.returncode, expected, "Stdout: {}; Stderr: {}".format(out, err)
            )
            return (out, err)
        else:
            self.assertEqual(
                p.returncode,
                expected,
                "Expected return code: %s doesn't match actual "
                "return code of: %s" % (expected, p.returncode),
            )
            output_stream = io.BytesIO(out)
            stream = subunit_lib.ByteStreamToStreamResult(output_stream)
            starts = testtools.StreamResult()
            summary = testtools.StreamSummary()
            tests = []

            def _add_dict(test):
                tests.append(test)

            outcomes = testtools.StreamToDict(functools.partial(_add_dict))
            result = testtools.CopyStreamResult([starts, outcomes, summary])
            result.startTestRun()
            try:
                stream.run(result)
            finally:
                result.stopTestRun()
            self.assertThat(len(tests), testtools.matchers.GreaterThan(0))
            return (out, err)

    def test_empty_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        self.assertRunExit("stestr --user-config=%s run" % path, 1)

    def test_empty_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        self.assertRunExit("stestr --user-config=%s run passing" % path, 0)

    def test_no_subunit_trace_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "no-subunit-trace": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        out, err = self.assertRunExit("stestr --user-config=%s run passing" % path, 0)
        out = str(out)
        self.assertIn("PASSED (id=0)", out)
        self.assertNotIn("Totals", out)
        self.assertNotIn("Worker Balance", out)
        self.assertNotIn("Sum of execute time for each test:", out)
        self.assertNotIn("Runtime (s)", out)

    def test_no_subunit_trace_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "no-subunit-trace": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        out, err = self.assertRunExit("stestr --user-config=%s run" % path, 1)
        out = str(out)
        self.assertIn("FAILED (id=0, failures=2)", out)
        self.assertNotIn("Totals", out)
        self.assertNotIn("Worker Balance", out)
        self.assertNotIn("Sum of execute time for each test:", out)

    def test_no_subunit_trace_config_file_force_subunit_trace(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "no-subunit-trace": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        out, err = self.assertRunExit(
            "stestr --user-config=%s run --force-subunit-trace passing" % path, 0
        )
        out = str(out)
        self.assertNotIn("PASSED (id=0)", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    def test_abbreviate_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "abbreviate": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        out, err = self.assertRunExit("stestr --user-config=%s run passing" % path, 0)
        out = str(out)
        self.assertIn("..", out)
        self.assertNotIn("PASSED (id=0)", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    def test_abbreviate_config_file_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "abbreviate": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        # NOTE(mtreinish): Running serially here to ensure a consistent
        # execution order for confirming the abbreviated output.
        out, err = self.assertRunExit("stestr --user-config=%s run --serial" % path, 1)
        out = str(out)
        self.assertIn("FF..", out)
        self.assertNotIn("FAILED (id=0, failures=2)", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    def test_no_subunit_trace_slowest_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "no-subunit-trace": True,
                        "slowest": True,
                    }
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        out, err = self.assertRunExit("stestr --user-config=%s run passing" % path, 0)
        out = str(out)
        self.assertIn("PASSED (id=0)", out)
        self.assertNotIn("Totals", out)
        self.assertNotIn("Worker Balance", out)
        self.assertNotIn("Sum of execute time for each test:", out)
        self.assertIn("Runtime (s)", out)

    def test_failing_list_config_file(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "no-subunit-trace": True,
                        "slowest": True,
                    },
                    "failing": {"list": True},
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        self.assertRunExit("stestr --user-config=%s run" % path, 1)
        out, err = self.assertRunExit("stestr --user-config=%s failing" % path, 1)
        out = str(out)
        self.assertNotIn("FAILED (id=0, failures=2)", out)
        self.assertNotIn("FAIL:", out)
        self.assertIn("tests.test_failing.FakeTestClass.test_pass", out)
        self.assertIn("tests.test_failing.FakeTestClass.test_pass_list", out)

    def test_no_subunit_trace_last_config_file_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "slowest": True,
                    },
                    "failing": {"list": True},
                    "last": {
                        "no-subunit-trace": True,
                    },
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        run_out, run_err = self.assertRunExit(
            "stestr --user-config=%s run passing" % path, 0
        )
        out, err = self.assertRunExit("stestr --user-config=%s last" % path, 0)
        run_out = str(run_out)
        out = str(out)
        self.assertIn("PASSED (id=0)", out)
        self.assertNotIn("Totals", out)
        self.assertNotIn("Worker Balance", out)
        self.assertNotIn("Sum of execute time for each test:", out)
        self.assertNotIn("Runtime (s)", out)
        self.assertIn("Totals", run_out)
        self.assertIn("Worker Balance", run_out)
        self.assertIn("Sum of execute time for each test:", run_out)
        self.assertIn("Runtime (s)", run_out)

    def test_subunit_trace_load_from_config_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "slowest": True,
                    },
                    "failing": {"list": True},
                    "last": {
                        "no-subunit-trace": True,
                    },
                    "load": {
                        "subunit-trace": True,
                    },
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))

        self.assertRunExit("stestr --user-config=%s run passing" % path, 0)
        stream = self._get_cmd_stdout("stestr --user-config=%s last --subunit" % path)[
            0
        ]
        out, err = self.assertRunExit(
            "stestr --user-config=%s load" % path, 0, stdin=stream
        )
        out = str(out)
        self.assertNotIn("PASSED (id=0)", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    def test_subunit_trace_load_from_config_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "slowest": True,
                    },
                    "failing": {"list": True},
                    "last": {
                        "no-subunit-trace": True,
                    },
                    "load": {
                        "subunit-trace": True,
                    },
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))

        self.assertRunExit("stestr --user-config=%s run" % path, 1)
        stream = self._get_cmd_stdout("stestr --user-config=%s last --subunit" % path)[
            0
        ]
        out, err = self.assertRunExit(
            "stestr --user-config=%s load" % path, 0, stdin=stream
        )
        out = str(out)
        self.assertNotIn("FAILED (id=0, failures=2)", out)
        self.assertNotIn("FF..", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    @testtools.skip("Abbreviated output not displaying")
    def test_abbreviate_load_from_config_passing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "slowest": True,
                    },
                    "failing": {"list": True},
                    "last": {
                        "no-subunit-trace": True,
                    },
                    "load": {
                        "abbreviate": True,
                    },
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))

        self.assertRunExit("stestr --user-config=%s run passing" % path, 0)
        stream = self._get_cmd_stdout("stestr --user-config=%s last --subunit" % path)[
            0
        ]
        out, err = self.assertRunExit(
            "stestr --user-config=%s load" % path, 0, stdin=stream
        )
        out = str(out)
        self.assertNotIn("PASSED (id=0)", out)
        self.assertIn("..", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    @testtools.skip("Abbreviated output not displaying")
    def test_abbreviate_load_from_config_failing(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        conf_file = os.fdopen(fd, "wb", 0)
        self.addCleanup(conf_file.close)
        contents = str(
            yaml.dump(
                {
                    "run": {
                        "slowest": True,
                    },
                    "failing": {"list": True},
                    "last": {
                        "no-subunit-trace": True,
                    },
                    "load": {
                        "abbreviate": True,
                    },
                },
                default_flow_style=False,
            )
        )
        conf_file.write(contents.encode("utf-8"))
        # NOTE(mtreinish): Running serially here to ensure a consistent
        # execution order for confirming the abbreviated output.
        self.assertRunExit("stestr --user-config=%s run --serial" % path, 1)
        stream = self._get_cmd_stdout("stestr --user-config=%s last --subunit" % path)[
            0
        ]
        out, err = self.assertRunExit(
            "stestr --user-config=%s load" % path, 0, stdin=stream
        )
        out = str(out)
        self.assertNotIn("FAILED (id=0, failures=2)", out)
        self.assertIn("FF..", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io

from stestr import output
from stestr.tests import base


class TestOutput(base.TestCase):
    def test_output_table(self):
        table = [
            ["Header 1", "Header 2", "Header 999"],
            [1, "0000000002", "foo"],
            ["bar", 6, "This is a content."],
        ]
        expected = (
            "Header 1  Header 2    Header 999\n"
            "--------  ----------  ------------------\n"
            "1         0000000002  foo\n"
            "bar       6           This is a content.\n"
        )
        with io.StringIO() as f:
            output.output_table(table, f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_tests(self):
        class Test:
            def __init__(self, i):
                self.i = i

            def id(self):
                return self.i

        tests = [Test("a"), Test("b"), Test("foo")]
        expected = "a\nb\nfoo\n"
        with io.StringIO() as f:
            output.output_tests(tests, f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_summary_passed(self):
        expected = (
            "Ran 10 (+5) tests in 1.100s (+0.100s)\n"
            "PASSED (id=99 (+1), id=100 (+2))\n"
        )
        with io.StringIO() as f:
            output.output_summary(
                successful=True,
                tests=10,
                tests_delta=5,
                time=1.1,
                time_delta=0.1,
                values=[("id", 99, 1), ("id", "100", 2)],
                output=f,
            )
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_summary_failed(self):
        expected = (
            "Ran 10 (+5) tests in 1.100s (+0.100s)\n"
            "FAILED (id=99 (+1), id=100 (+2))\n"
        )
        with io.StringIO() as f:
            output.output_summary(
                successful=False,
                tests=10,
                tests_delta=5,
                time=1.1,
                time_delta=0.1,
                values=[("id", 99, 1), ("id", "100", 2)],
                output=f,
            )
            actual = f.getvalue()
            self.assertEqual(expected, actual)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import re
from unittest import mock

from stestr import selection
from stestr.tests import base


class TestSelection(base.TestCase):
    def test_filter_tests_no_filter(self):
        test_list = ["a", "b", "c"]
        result = selection.filter_tests(None, test_list)
        self.assertEqual(test_list, result)

    def test_filter_tests(self):
        test_list = ["a", "b", "c"]
        result = selection.filter_tests(["a"], test_list)
        self.assertEqual(["a"], result)

    def test_filter_invalid_regex(self):
        test_list = ["a", "b", "c"]
        with mock.patch("sys.exit", side_effect=ImportError) as mock_exit:
            self.assertRaises(
                ImportError,
                selection.filter_tests,
                ["fake_regex_with_bad_part[The-BAD-part]"],
                test_list,
            )
            mock_exit.assert_called_once_with(5)


class TestExclusionReader(base.TestCase):
    def test_exclusion_reader(self):
        exclude_list = io.StringIO()
        for i in range(4):
            exclude_list.write("fake_regex_%s\n" % i)
            exclude_list.write("fake_regex_with_note_%s # note\n" % i)
        exclude_list.seek(0)
        with mock.patch("builtins.open", return_value=exclude_list):
            result = selection.exclusion_reader("fake_path")
        self.assertEqual(2 * 4, len(result))
        note_cnt = 0
        # not assuming ordering, mainly just testing the type
        for r in result:
            self.assertEqual(r[2], [])
            if r[1] == "note":
                note_cnt += 1
            self.assertIn("search", dir(r[0]))  # like a compiled regexp
        self.assertEqual(note_cnt, 4)

    def test_invalid_regex(self):
        exclude_list = io.StringIO()
        exclude_list.write("fake_regex_with_bad_part[The-BAD-part]")
        exclude_list.seek(0)
        with mock.patch("builtins.open", return_value=exclude_list):
            with mock.patch("sys.exit") as mock_exit:
                selection.exclusion_reader("fake_path")
                mock_exit.assert_called_once_with(5)


class TestConstructList(base.TestCase):
    def test_simple_re(self):
        test_lists = ["fake_test(scen)[tag,bar])", "fake_test(scen)[egg,foo])"]
        result = selection.construct_list(test_lists, regexes=["foo"])
        self.assertEqual(list(result), ["fake_test(scen)[egg,foo])"])

    def test_simple_exclusion_re(self):
        test_lists = ["fake_test(scen)[tag,bar])", "fake_test(scen)[egg,foo])"]
        result = selection.construct_list(test_lists, exclude_regex="foo")
        self.assertEqual(list(result), ["fake_test(scen)[tag,bar])"])

    def test_invalid_exclusion_re(self):
        test_lists = ["fake_test(scen)[tag,bar])", "fake_test(scen)[egg,foo])"]
        invalid_regex = "fake_regex_with_bad_part[The-BAD-part]"
        with mock.patch("sys.exit", side_effect=ImportError) as exit_mock:
            self.assertRaises(
                ImportError,
                selection.construct_list,
                test_lists,
                exclude_regex=invalid_regex,
            )
            exit_mock.assert_called_once_with(5)

    def test_exclusion_list(self):
        exclude_list = [(re.compile("foo"), "foo not liked", [])]
        test_lists = ["fake_test(scen)[tag,bar])", "fake_test(scen)[egg,foo])"]
        with mock.patch("stestr.selection.exclusion_reader", return_value=exclude_list):
            result = selection.construct_list(
                test_lists, exclude_list="file", regexes=["fake_test"]
            )
        self.assertEqual(list(result), ["fake_test(scen)[tag,bar])"])

    def test_inclusion_list(self):
        include_list = [re.compile("fake_test1"), re.compile("fake_test2")]
        test_lists = ["fake_test1[tg]", "fake_test2[tg]", "fake_test3[tg]"]
        include_getter = "stestr.selection._get_regex_from_include_list"
        with mock.patch(include_getter, return_value=include_list):
            result = selection.construct_list(test_lists, include_list="file")
        self.assertEqual(set(result), {"fake_test1[tg]", "fake_test2[tg]"})

    def test_inclusion_list_invalid_regex(self):
        include_list = io.StringIO()
        include_list.write("fake_regex_with_bad_part[The-BAD-part]")
        include_list.seek(0)
        with mock.patch("builtins.open", return_value=include_list):
            with mock.patch("sys.exit") as mock_exit:
                selection._get_regex_from_include_list("fake_path")
                mock_exit.assert_called_once_with(5)

    def test_inclusion_exclusion_list_re(self):
        include_list = [re.compile("fake_test1"), re.compile("fake_test2")]
        test_lists = [
            "fake_test1[tg]",
            "fake_test2[spam]",
            "fake_test3[tg,foo]",
            "fake_test4[spam]",
        ]
        exclude_list = [(re.compile("spam"), "spam not liked", [])]
        include_getter = "stestr.selection._get_regex_from_include_list"
        with mock.patch(include_getter, return_value=include_list):
            with mock.patch(
                "stestr.selection.exclusion_reader", return_value=exclude_list
            ):
                result = selection.construct_list(
                    test_lists,
                    exclude_list="exclude_file",
                    include_list="include_file",
                    regexes=["foo"],
                )
        self.assertEqual(set(result), {"fake_test1[tg]", "fake_test3[tg,foo]"})

    def test_overlapping_exclude_regex(self):

        exclude_list = [
            (re.compile("compute.test_keypairs.KeypairsTestV210"), "", []),
            (re.compile("compute.test_keypairs.KeypairsTestV21"), "", []),
        ]
        test_lists = [
            "compute.test_keypairs.KeypairsTestV210.test_create_keypair",
            "compute.test_keypairs.KeypairsTestV21.test_create_keypair",
            "compute.test_fake.FakeTest.test_fake_test",
        ]
        with mock.patch("stestr.selection.exclusion_reader", return_value=exclude_list):
            result = selection.construct_list(
                test_lists, exclude_list="file", regexes=["fake_test"]
            )
        self.assertEqual(list(result), ["compute.test_fake.FakeTest.test_fake_test"])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil
import subprocess
import tempfile

from stestr.tests import base


class TestBisectReturnCodes(base.TestCase):
    def setUp(self):
        super().setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix="stestr-unit")
        self.addCleanup(shutil.rmtree, self.directory)
        self.test_dir = os.path.join(self.directory, "tests")
        os.mkdir(self.test_dir)
        # Setup Test files
        self.testr_conf_file = os.path.join(self.directory, ".stestr.conf")
        self.setup_cfg_file = os.path.join(self.directory, "setup.cfg")
        self.init_file = os.path.join(self.test_dir, "__init__.py")
        self.setup_py = os.path.join(self.directory, "setup.py")
        self.user_config = os.path.join(self.directory, "stestr.yaml")
        shutil.copy("stestr/tests/files/testr-conf", self.testr_conf_file)
        shutil.copy("stestr/tests/files/setup.py", self.setup_py)
        shutil.copy("stestr/tests/files/setup.cfg", self.setup_cfg_file)
        shutil.copy("stestr/tests/files/__init__.py", self.init_file)
        shutil.copy("stestr/tests/files/stestr.yaml", self.user_config)

        # Move around the test code
        self.serial_fail_file = os.path.join(self.test_dir, "test_serial_fails.py")
        shutil.copy(
            "stestr/tests/files/bisect-fail-serial-tests", self.serial_fail_file
        )

        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, os.path.abspath(os.curdir))
        os.chdir(self.directory)
        subprocess.call("stestr init", shell=True)

    def test_bisect_serial_fail_detected(self):
        p = subprocess.Popen(
            "stestr run --serial",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        self.assertEqual(
            1,
            p.returncode,
            "stestr run returned an unexpected return code"
            "Stdout: %s\nStderr: %s" % (out, err),
        )
        p_analyze = subprocess.Popen(
            "stestr --user-config stestr.yaml run --analyze-isolation",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p_analyze.communicate()
        out = out.decode("utf-8")
        # For debugging potential failures
        lines = str(out.rstrip()).splitlines()
        self.assertEqual(
            3,
            p_analyze.returncode,
            "Analyze isolation returned an unexpected return code"
            "\nStdout: %s\nStderr: %s" % (out, err),
        )
        last_line = (
            "tests.test_serial_fails.TestFakeClass.test_B  "
            "tests.test_serial_fails.TestFakeClass.test_A"
        )
        self.assertEqual(last_line, lines[-1])



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


def find_and_remove(suffix=".pyc"):
    for root, dirs, files in os.walk("."):
        for file in files:
            target = os.path.join(root, file)
            if os.path.isfile(target) and target.endswith(suffix):
                os.remove(target)


if __name__ == "__main__":
    find_and_remove()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import subprocess
from unittest import mock

from stestr import test_processor
from stestr.tests import base


class TestTestProcessorFixture(base.TestCase):
    def setUp(self):
        super().setUp()
        self._fixture = test_processor.TestProcessorFixture(
            mock.sentinel.test_ids,
            mock.sentinel.options,
            mock.sentinel.cmd_template,
            mock.sentinel.listopt,
            mock.sentinel.idoption,
            mock.sentinel.repository,
        )

    @mock.patch.object(subprocess, "Popen")
    @mock.patch.object(test_processor, "sys")
    def _check_start_process(
        self, mock_sys, mock_Popen, platform="win32", expected_fn=None
    ):
        mock_sys.platform = platform

        self._fixture._start_process(mock.sentinel.cmd)

        mock_Popen.assert_called_once_with(
            mock.sentinel.cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            preexec_fn=expected_fn,
        )

    def test_start_process_win32(self):
        self._check_start_process()

    def test_start_process_linux(self):
        self._check_start_process(
            platform="linux2", expected_fn=self._fixture._clear_SIGPIPE
        )



# Copyright 2015 NEC Corporation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# Colorizer Code is borrowed from Twisted:
# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files (the
#    "Software"), to deal in the Software without restriction, including
#    without limitation the rights to use, copy, modify, merge, publish,
#    distribute, sublicense, and/or sell copies of the Software, and to
#    permit persons to whom the Software is furnished to do so, subject to
#    the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
#    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
#    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import sys


class AnsiColorizer:
    """A colorizer is an object that loosely wraps around a stream

    allowing callers to write text to the stream in a particular color.

    Colorizer classes must implement C{supported()} and C{write(text, color)}.
    """

    _colors = dict(
        black=30, red=31, green=32, yellow=33, blue=34, magenta=35, cyan=36, white=37
    )

    def __init__(self, stream):
        self.stream = stream

    @classmethod
    def supported(cls, stream=sys.stdout):
        """Check the current platform supports coloring terminal output

        A class method that returns True if the current platform supports
        coloring terminal output using this method. Returns False otherwise.
        """
        if not stream.isatty():
            return False  # auto color only on TTYs
        try:
            import curses
        except ImportError:
            return False
        else:
            try:
                try:
                    return curses.tigetnum("colors") > 2
                except curses.error:
                    curses.setupterm()
                    return curses.tigetnum("colors") > 2
            except Exception:
                # guess false in case of error
                return False

    def write(self, text, color):
        """Write the given text to the stream in the given color.

        @param text: Text to be written to the stream.

        @param color: A string label for a color. e.g. 'red', 'white'.
        """
        color = self._colors[color]
        self.stream.write("\x1b[{};1m{}\x1b[0m".format(color, text))


class NullColorizer:
    """See _AnsiColorizer docstring."""

    def __init__(self, stream):
        self.stream = stream

    @classmethod
    def supported(cls, stream=sys.stdout):
        return True

    def write(self, text, color):
        self.stream.write(text)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""List the tests from a project and show them."""

from io import BytesIO
import sys

from cliff import command

from stestr import config_file
from stestr import output


class List(command.Command):
    """List the tests for a project.

    You can use a filter just like with the run command to see exactly what
    tests match.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "filters",
            nargs="*",
            default=None,
            help="A list of string regex filters to initially "
            "apply on the test list. Tests that match any of "
            "the regexes will be used. (assuming any other "
            "filtering specified also uses it)",
        )
        parser.add_argument(
            "--exclude-list",
            "-e",
            default=None,
            dest="exclude_list",
            help="Path to an exclusion list file, this file "
            "contains a separate regex exclude on each "
            "newline",
        )
        parser.add_argument(
            "--include-list",
            "-i",
            default=None,
            dest="include_list",
            help="Path to an inclusion list file, this file "
            "contains a separate regex on each newline.",
        )
        parser.add_argument(
            "--exclude-regex",
            "-E",
            default=None,
            dest="exclude_regex",
            help="Test rejection regex. If a test cases name "
            "matches on re.search() operation , "
            "it will be removed from the final test list. "
            "Effectively the exclusion-regexp is added to "
            "exclusion regexp list, but you do need to edit a "
            "file. The exclusion filtering happens after the "
            "initial safe list selection, which by default is "
            "everything.",
        )
        return parser

    def take_action(self, parsed_args):
        args = parsed_args
        filters = parsed_args.filters or None
        return list_command(
            config=self.app_args.config,
            repo_url=self.app_args.repo_url,
            group_regex=self.app_args.group_regex,
            test_path=self.app_args.test_path,
            top_dir=self.app_args.top_dir,
            exclude_list=args.exclude_list,
            include_list=args.include_list,
            exclude_regex=args.exclude_regex,
            filters=filters,
        )


def list_command(
    config=config_file.TestrConf.DEFAULT_CONFIG_FILENAME,
    repo_url=None,
    test_path=None,
    top_dir=None,
    group_regex=None,
    exclude_list=None,
    include_list=None,
    exclude_regex=None,
    filters=None,
    stdout=sys.stdout,
):
    """Print a list of test_ids for a project

    This function will print the test_ids for tests in a project. You can
    filter the output just like with the run command to see exactly what
    will be run.

    :param str config: The path to the stestr config file. Must be a string.
    :param str repo_url: The url of the repository to use.
    :param str test_path: Set the test path to use for unittest discovery.
        If both this and the corresponding config file option are set, this
        value will be used.
    :param str top_dir: The top dir to use for unittest discovery. This takes
        precedence over the value in the config file. (if one is present in
        the config file)
    :param str group_regex: Set a group regex to use for grouping tests
        together in the stestr scheduler. If both this and the corresponding
        config file option are set this value will be used.
    :param str exclude_list: Path to an exclusion list file, this file
        contains a separate regex exclude on each newline.
    :param str include_list: Path to an inclusion list file, this file
        contains a separate regex on each newline.
    :param str exclude_regex: Test rejection regex. If a test cases name
        matches on re.search() operation, it will be removed from the final
        test list.
    :param list filters: A list of string regex filters to initially apply on
        the test list. Tests that match any of the regexes will be used.
        (assuming any other filtering specified also uses it)
    :param file stdout: The output file to write all output to. By default
        this is sys.stdout

    """
    ids = None
    conf = config_file.TestrConf.load_from_file(config)
    cmd = conf.get_run_command(
        regexes=filters,
        repo_url=repo_url,
        group_regex=group_regex,
        exclude_list=exclude_list,
        include_list=include_list,
        exclude_regex=exclude_regex,
        test_path=test_path,
        top_dir=top_dir,
    )
    not_filtered = (
        filters is None
        and include_list is None
        and exclude_list is None
        and exclude_regex is None
    )
    try:
        cmd.setUp()
        # List tests if the fixture has not already needed to to filter.
        if not_filtered:
            ids = cmd.list_tests()
        else:
            ids = cmd.test_ids
        stream = BytesIO()
        for id in ids:
            stream.write(("%s\n" % id).encode("utf8"))
        stream.seek(0)
        output.output_stream(stream, output=stdout)
        return 0
    finally:
        cmd.cleanUp()



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import os
import sys


if not os.path.isfile(".testr.conf"):
    sys.exit("Testr config file not found")

with open(".testr.conf", "r") as testr_conf_file:
    config = configparser.ConfigParser()
    config.readfp(testr_conf_file)

    test_command = config.get("DEFAULT", "test_command")
    group_regex = None
    if config.has_option("DEFAULT", "group_regex"):
        group_regex = config.get("DEFAULT", "group_regex")

top_dir = None
test_dir = None
for line in test_command.split("\n"):
    if "subunit.run discover" in line:
        command_parts = line.split(" ")
        top_dir_present = "-t" in line
        for idx, val in enumerate(command_parts):
            if top_dir_present:
                if val == "-t":
                    top_dir = command_parts[idx + 1]
                    test_dir = command_parts[idx + 2]
            else:
                if val == "discover":
                    test_dir = command_parts[idx + 1]

with open(".stestr.conf", "w") as stestr_conf_file:
    stestr_conf_file.write("[DEFAULT]\n")
    stestr_conf_file.write("test_path=%s\n" % test_dir)
    if top_dir:
        stestr_conf_file.write("top_dir=%s\n" % top_dir)
    if group_regex:
        stestr_conf_file.write("group_regex=%s\n" % group_regex)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

import voluptuous as vp
import yaml


def get_user_config(path=None):
    if not path:
        home_dir = os.path.expanduser("~")
        path = os.path.join(home_dir, ".stestr.yaml")
        if not os.path.isfile(path):
            path = os.path.join(os.path.join(home_dir, ".config"), "stestr.yaml")
            if not os.path.isfile(path):
                path = None
        if not path:
            return None
    else:
        if not os.path.isfile(path):
            msg = "The specified stestr user config is not a valid path"
            sys.exit(msg)

    return UserConfig(path)


class UserConfig:
    def __init__(self, path):
        self.schema = vp.Schema(
            {
                vp.Optional("run"): {
                    vp.Optional("concurrency"): int,
                    vp.Optional("random"): bool,
                    vp.Optional("no-subunit-trace"): bool,
                    vp.Optional("color"): bool,
                    vp.Optional("abbreviate"): bool,
                    vp.Optional("slowest"): bool,
                    vp.Optional("suppress-attachments"): bool,
                    vp.Optional("all-attachments"): bool,
                },
                vp.Optional("failing"): {
                    vp.Optional("list"): bool,
                },
                vp.Optional("last"): {
                    vp.Optional("no-subunit-trace"): bool,
                    vp.Optional("color"): bool,
                    vp.Optional("suppress-attachments"): bool,
                    vp.Optional("all-attachments"): bool,
                },
                vp.Optional("load"): {
                    vp.Optional("force-init"): bool,
                    vp.Optional("subunit-trace"): bool,
                    vp.Optional("color"): bool,
                    vp.Optional("abbreviate"): bool,
                    vp.Optional("suppress-attachments"): bool,
                    vp.Optional("all-attachments"): bool,
                },
                vp.Optional("history-list"): {
                    vp.Optional("show-metadata"): bool,
                },
                vp.Optional("history-show"): {
                    vp.Optional("no-subunit-trace"): bool,
                    vp.Optional("color"): bool,
                    vp.Optional("suppress-attachments"): bool,
                    vp.Optional("all-attachments"): bool,
                },
            }
        )
        with open(path) as fd:
            self.config = yaml.safe_load(fd.read())
        if self.config is None:
            self.config = {}
        try:
            self.schema(self.config)
        except vp.MultipleInvalid as e:
            msg = "Provided user config file {} is invalid " "because:\n{}".format(
                path, str(e)
            )
            sys.exit(msg)

    @property
    def run(self):
        return self.config.get("run")

    @property
    def failing(self):
        return self.config.get("failing")

    @property
    def last(self):
        return self.config.get("last")

    @property
    def load(self):
        return self.config.get("load")

    @property
    def history_list(self):
        return self.config.get("history-list")

    @property
    def history_show(self):
        return self.config.get("history-show")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""In memory storage of test results."""

from collections import OrderedDict
from io import BytesIO
from operator import methodcaller

import subunit
import testtools

from stestr.repository import abstract as repository


class RepositoryFactory(repository.AbstractRepositoryFactory):
    """A factory that can initialise and open memory repositories.

    This is used for testing where a repository may be created and later
    opened, but tests should not see each others repositories.
    """

    def __init__(self):
        self.repos = {}

    def initialise(self, url):
        self.repos[url] = Repository()
        return self.repos[url]

    def open(self, url):
        try:
            return self.repos[url]
        except KeyError:
            raise repository.RepositoryNotFound(url)


class Repository(repository.AbstractRepository):
    """In memory storage of test results."""

    def __init__(self):
        # Test runs:
        self._runs = []
        self._failing = OrderedDict()  # id -> test
        self._times = {}  # id -> duration

    def count(self):
        return len(self._runs)

    def get_failing(self):
        return _Failures(self)

    def get_test_run(self, run_id):
        if run_id < 0:
            raise KeyError("No such run.")
        return self._runs[run_id]

    def get_run_ids(self):
        return list(self._runs.keys())

    def remove_run_id(self, run_id):
        if run_id not in self._runs:
            raise KeyError("No run %s in repository" % run_id)
        del self._run[run_id]

    def latest_id(self):
        result = self.count() - 1
        if result < 0:
            raise KeyError("No tests in repository")
        return result

    def _get_inserter(self, partial, run_id=None, metadata=None):
        return _Inserter(self, partial, run_id, metadata)

    def _get_test_times(self, test_ids):
        result = {}
        for test_id in test_ids:
            duration = self._times.get(test_id, None)
            if duration is not None:
                result[test_id] = duration
        return result


# XXX: Too much duplication between this and _Inserter
class _Failures(repository.AbstractTestRun):
    """Report on failures from a memory repository."""

    def __init__(self, repository):
        self._repository = repository

    def get_id(self):
        return None

    def get_subunit_stream(self):
        result = BytesIO()
        serialiser = subunit.v2.StreamResultToBytes(result)
        serialiser = testtools.ExtendedToStreamDecorator(serialiser)
        serialiser.startTestRun()
        try:
            self.run(serialiser)
        finally:
            serialiser.stopTestRun()
        result.seek(0)
        return result

    def get_test(self):
        def wrap_result(result):
            # Wrap in a router to mask out startTestRun/stopTestRun from the
            # ExtendedToStreamDecorator.
            result = testtools.StreamResultRouter(result, do_start_stop_run=False)
            # Wrap that in ExtendedToStreamDecorator to convert v1 calls to
            # StreamResult.
            return testtools.ExtendedToStreamDecorator(result)

        return testtools.DecorateTestCaseResult(
            self, wrap_result, methodcaller("startTestRun"), methodcaller("stopTestRun")
        )

    def run(self, result):
        # Speaks original V1 protocol.
        for case in self._repository._failing.values():
            case.run(result)


class _Inserter(repository.AbstractTestRun):
    """Insert test results into a memory repository."""

    def __init__(self, repository, partial, run_id=None, metadata=None):
        self._repository = repository
        self._partial = partial
        self._tests = []
        # Subunit V2 stream for get_subunit_stream
        self._subunit = None
        self._run_id = run_id
        self._metadata = metadata

    def startTestRun(self):
        self._subunit = BytesIO()
        serialiser = subunit.v2.StreamResultToBytes(self._subunit)
        self._hook = testtools.CopyStreamResult(
            [testtools.StreamToDict(self._handle_test), serialiser]
        )
        self._hook.startTestRun()

    def _handle_test(self, test_dict):
        self._tests.append(test_dict)
        start, stop = test_dict["timestamps"]
        if test_dict["status"] == "exists" or None in (start, stop):
            return
        duration_delta = stop - start
        duration_seconds = (
            duration_delta.microseconds
            + (duration_delta.seconds + duration_delta.days * 24 * 3600) * 10**6
        ) / 10.0**6
        self._repository._times[test_dict["id"]] = duration_seconds

    def stopTestRun(self):
        self._hook.stopTestRun()
        self._repository._runs.append(self)
        if not self._run_id:
            self._run_id = len(self._repository._runs) - 1
        if not self._partial:
            self._repository._failing = OrderedDict()
        for test_dict in self._tests:
            test_id = test_dict["id"]
            if test_dict["status"] == "fail":
                case = testtools.testresult.real.test_dict_to_case(test_dict)
                self._repository._failing[test_id] = case
            else:
                self._repository._failing.pop(test_id, None)
        return self._run_id

    def status(self, *args, **kwargs):
        self._hook.status(*args, **kwargs)

    def get_id(self):
        return self._run_id

    def get_subunit_stream(self):
        self._subunit.seek(0)
        return self._subunit

    def get_test(self):
        def wrap_result(result):
            # Wrap in a router to mask out startTestRun/stopTestRun from the
            # ExtendedToStreamDecorator.
            result = testtools.StreamResultRouter(result, do_start_stop_run=False)
            # Wrap that in ExtendedToStreamDecorator to convert v1 calls to
            # StreamResult.
            return testtools.ExtendedToStreamDecorator(result)

        return testtools.DecorateTestCaseResult(
            self, wrap_result, methodcaller("startTestRun"), methodcaller("stopTestRun")
        )

    def run(self, result):
        # Speaks original.
        for test_dict in self._tests:
            case = testtools.testresult.real.test_dict_to_case(test_dict)
            case.run(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Run a projects tests and load them into stestr."""

import errno
import functools
import io
import os
import os.path
import subprocess
import sys

from cliff import command
import subunit
import testtools

from stestr import bisect_tests
from stestr.commands import load
from stestr.commands import slowest
from stestr import config_file
from stestr import output
from stestr.repository import abstract as repository
from stestr.repository import util
from stestr import results
from stestr.subunit_runner import program
from stestr.subunit_runner import run as subunit_run
from stestr.testlist import parse_list
from stestr import user_config


def _to_int(possible, default=0, out=sys.stderr):
    try:
        i = int(possible)
    except (ValueError, TypeError):
        i = default
        msg = 'Unable to convert "%s" to an integer.  Using %d.\n' % (possible, default)
        out.write(str(msg))
    return i


class Run(command.Command):
    """Run the tests for a project and store them into the repository.

    Without --subunit, the process exit code will be non-zero if the test
    run was not successful. However, with --subunit, the process exit code
    is non-zero only if the subunit stream could not be generated
    successfully. The test results and run status are included in the
    subunit stream, so the stream should be used to determining the result
    of the run instead of the exit code when using the --subunit flag.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "filters",
            nargs="*",
            default=None,
            help="A list of string regex filters to initially "
            "apply on the test list. Tests that match any of "
            "the regexes will be used. (assuming any other "
            "filtering specified also uses it)",
        )
        parser.add_argument(
            "--failing",
            action="store_true",
            default=False,
            help="Run only tests known to be failing.",
        )
        parser.add_argument(
            "--serial",
            action="store_true",
            default=False,
            help="Run tests in a serial process.",
        )
        parser.add_argument(
            "--concurrency",
            action="store",
            default=None,
            type=int,
            help="How many processes to use. The default (0) "
            "autodetects your CPU count.",
        )
        parser.add_argument(
            "--load-list", default=None, help="Only run tests listed in the named file."
        ),
        parser.add_argument(
            "--subunit",
            action="store_true",
            default=False,
            help="Display results in subunit format.",
        )
        parser.add_argument(
            "--until-failure",
            action="store_true",
            default=False,
            help="Repeat the run again and again until " "failure occurs.",
        )
        parser.add_argument(
            "--analyze-isolation",
            action="store_true",
            default=False,
            help="Search the last test run for 2-test test " "isolation interactions.",
        )
        parser.add_argument(
            "--isolated",
            action="store_true",
            default=False,
            help="Run each test id in a separate test runner.",
        )
        parser.add_argument(
            "--worker-file",
            action="store",
            default=None,
            dest="worker_path",
            help="Optional path of a manual worker grouping " "file to use for the run",
        )
        parser.add_argument(
            "--exclude-list",
            "-e",
            default=None,
            dest="exclude_list",
            help="Path to an exclusion list file, this file "
            "contains a separate regex exclude on each "
            "newline",
        )
        parser.add_argument(
            "--include-list",
            "-i",
            default=None,
            dest="include_list",
            help="Path to an inclusion list file, this file "
            "contains a separate regex on each newline.",
        )
        parser.add_argument(
            "--exclude-regex",
            "-E",
            default=None,
            dest="exclude_regex",
            help="Test rejection regex. If a test cases name "
            "matches on re.search() operation , "
            "it will be removed from the final test list. "
            "Effectively the exclusion-regexp is added to "
            "exclusion regexp list, but you do need to edit a "
            "file. The exclusion filtering happens after the "
            "initial safe list selection, which by default is "
            "everything.",
        )
        parser.add_argument(
            "--no-discover",
            "-n",
            default=None,
            metavar="TEST_ID",
            help="Takes in a single test to bypasses test "
            "discover and just execute the test specified. A "
            "file may be used in place of a test name.",
        )
        parser.add_argument(
            "--pdb",
            default=None,
            metavar="TEST_ID",
            help="Run a single test id with the intent of "
            "using pdb. This does not launch any separate "
            "processes to ensure pdb works as expected. It "
            "will bypass test discovery and just execute the "
            "test specified. A file may be used in place of a "
            "test name.",
        )
        parser.add_argument(
            "--random",
            action="store_true",
            default=False,
            help="Randomize the test order after they are "
            "partitioned into separate workers",
        )
        parser.add_argument(
            "--combine",
            action="store_true",
            default=False,
            help="Combine the results from the test run with "
            "the last run in the repository",
        )
        parser.add_argument(
            "--no-subunit-trace",
            action="store_true",
            default=False,
            help="Disable the default subunit-trace output " "filter",
        )
        parser.add_argument(
            "--force-subunit-trace",
            action="store_true",
            default=False,
            help="Force subunit-trace output regardless of any"
            "other options or config settings",
        )
        parser.add_argument(
            "--color",
            action="store_true",
            default=False,
            help="Enable color output in the subunit-trace "
            "output, if subunit-trace output is enabled. "
            "(this is the default). If subunit-trace is "
            "disable this does nothing.",
        )
        parser.add_argument(
            "--slowest",
            action="store_true",
            default=False,
            help="After the test run, print the slowest " "tests.",
        )
        parser.add_argument(
            "--abbreviate",
            action="store_true",
            dest="abbreviate",
            help="Print one character status for each test",
        )
        parser.add_argument(
            "--suppress-attachments",
            action="store_true",
            dest="suppress_attachments",
            help="If set do not print stdout or stderr "
            "attachment contents on a successful test "
            "execution",
        )
        parser.add_argument(
            "--all-attachments",
            action="store_true",
            dest="all_attachments",
            help="If set print all text attachment contents on"
            " a successful test execution",
        )
        parser.add_argument(
            "--show-binary-attachments",
            action="store_true",
            dest="show_binary_attachments",
            help="If set, show non-text attachments. This is "
            "generally only useful for debug purposes.",
        )
        return parser

    def take_action(self, parsed_args):
        user_conf = user_config.get_user_config(self.app_args.user_config)
        filters = parsed_args.filters
        args = parsed_args
        if args.suppress_attachments and args.all_attachments:
            msg = (
                "The --suppress-attachments and --all-attachments "
                "options are mutually exclusive, you can not use both "
                "at the same time"
            )
            print(msg)
            sys.exit(1)
        if getattr(user_conf, "run", False):
            if not user_conf.run.get("no-subunit-trace"):
                if not args.no_subunit_trace:
                    pretty_out = True
                else:
                    pretty_out = False
            else:
                pretty_out = False

            pretty_out = args.force_subunit_trace or pretty_out
            if args.concurrency is None:
                concurrency = user_conf.run.get("concurrency", 0)
            else:
                concurrency = args.concurrency
            random = args.random or user_conf.run.get("random", False)
            color = args.color or user_conf.run.get("color", False)
            abbreviate = args.abbreviate or user_conf.run.get("abbreviate", False)
            suppress_attachments_conf = user_conf.run.get("suppress-attachments", False)
            all_attachments_conf = user_conf.run.get("all-attachments", False)
            if not args.suppress_attachments and not args.all_attachments:
                suppress_attachments = suppress_attachments_conf
                all_attachments = all_attachments_conf
            elif args.suppress_attachments:
                all_attachments = False
                suppress_attachments = args.suppress_attachments
            elif args.all_attachments:
                suppress_attachments = False
                all_attachments = args.all_attachments
        else:
            pretty_out = args.force_subunit_trace or not args.no_subunit_trace
            concurrency = args.concurrency or 0
            random = args.random
            color = args.color
            abbreviate = args.abbreviate
            suppress_attachments = args.suppress_attachments
            all_attachments = args.all_attachments
        verbose_level = self.app.options.verbose_level
        stdout = open(os.devnull, "w") if verbose_level == 0 else sys.stdout
        # Make sure all (python) callers have provided an int()
        concurrency = _to_int(concurrency)
        if concurrency and concurrency < 0:
            msg = (
                "The provided concurrency value: %s is not valid. An "
                "integer >= 0 must be used.\n" % concurrency
            )
            stdout.write(msg)
            return 2
        result = run_command(
            config=self.app_args.config,
            repo_url=self.app_args.repo_url,
            test_path=self.app_args.test_path,
            top_dir=self.app_args.top_dir,
            group_regex=self.app_args.group_regex,
            failing=args.failing,
            serial=args.serial,
            concurrency=concurrency,
            load_list=args.load_list,
            subunit_out=args.subunit,
            until_failure=args.until_failure,
            analyze_isolation=args.analyze_isolation,
            isolated=args.isolated,
            worker_path=args.worker_path,
            exclude_list=args.exclude_list,
            include_list=args.include_list,
            exclude_regex=args.exclude_regex,
            no_discover=args.no_discover,
            random=random,
            combine=args.combine,
            filters=filters,
            pretty_out=pretty_out,
            color=color,
            stdout=stdout,
            abbreviate=abbreviate,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=args.show_binary_attachments,
            pdb=args.pdb,
        )

        # Always output slowest test info if requested, regardless of other
        # test run options
        user_slowest = False
        if getattr(user_conf, "run", False):
            user_slowest = user_conf.run.get("slowest", False)
        if args.slowest or user_slowest:
            slowest.slowest(repo_url=self.app_args.repo_url)

        return result


def _find_failing(repo):
    run = repo.get_failing()
    case = run.get_test()
    ids = []

    def gather_errors(test_dict):
        if test_dict["status"] == "fail":
            ids.append(test_dict["id"])

    result = testtools.StreamToDict(gather_errors)
    result.startTestRun()
    try:
        case.run(result)
    finally:
        result.stopTestRun()
    return ids


def run_command(
    config=config_file.TestrConf.DEFAULT_CONFIG_FILENAME,
    repo_url=None,
    test_path=None,
    top_dir=None,
    group_regex=None,
    failing=False,
    serial=False,
    concurrency=0,
    load_list=None,
    subunit_out=False,
    until_failure=False,
    analyze_isolation=False,
    isolated=False,
    worker_path=None,
    exclude_list=None,
    include_list=None,
    exclude_regex=None,
    no_discover=False,
    random=False,
    combine=False,
    filters=None,
    pretty_out=True,
    color=False,
    stdout=sys.stdout,
    abbreviate=False,
    suppress_attachments=False,
    all_attachments=False,
    show_binary_attachments=True,
    pdb=False,
):
    """Function to execute the run command

    This function implements the run command. It will run the tests specified
    in the parameters based on the provided config file and/or arguments
    specified in the way specified by the arguments. The results will be
    printed to STDOUT and loaded into the repository.

    :param str config: The path to the stestr config file. Must be a string.
    :param str repo_url: The url of the repository to use.
    :param str test_path: Set the test path to use for unittest discovery.
        If both this and the corresponding config file option are set, this
        value will be used.
    :param str top_dir: The top dir to use for unittest discovery. This takes
        precedence over the value in the config file. (if one is present in
        the config file)
    :param str group_regex: Set a group regex to use for grouping tests
        together in the stestr scheduler. If both this and the corresponding
        config file option are set this value will be used.
    :param bool failing: Run only tests known to be failing.
    :param bool serial: Run tests serially
    :param int concurrency: "How many processes to use. The default (0)
        autodetects your CPU count and uses that.
    :param str load_list: The path to a list of test_ids. If specified only
        tests listed in the named file will be run.
    :param bool subunit_out: Display results in subunit format.
    :param bool until_failure: Repeat the run again and again until failure
        occurs.
    :param bool analyze_isolation: Search the last test run for 2-test test
        isolation interactions.
    :param bool isolated: Run each test id in a separate test runner.
    :param str worker_path: Optional path of a manual worker grouping file
        to use for the run.
    :param str exclude_list: Path to an exclusion list file, this file
        contains a separate regex exclude on each newline.
    :param str include_list: Path to a inclusion list file, this file
        contains a separate regex on each newline.
    :param str exclude_regex: Test rejection regex. If a test cases name
        matches on re.search() operation, it will be removed from the final
        test list.
    :param str no_discover: Takes in a single test_id to bypasses test
        discover and just execute the test specified. A file name may be used
        in place of a test name.
    :param bool random: Randomize the test order after they are partitioned
        into separate workers
    :param bool combine: Combine the results from the test run with the
        last run in the repository
    :param list filters: A list of string regex filters to initially apply on
        the test list. Tests that match any of the regexes will be used.
        (assuming any other filtering specified also uses it)
    :param bool pretty_out: Use the subunit-trace output filter
    :param bool color: Enable colorized output in subunit-trace
    :param file stdout: The file object to write all output to. By default this
        is sys.stdout
    :param bool abbreviate: Use abbreviated output if set true
    :param bool suppress_attachments: When set true attachments subunit_trace
        will not print attachments on successful test execution.
    :param bool all_attachments: When set true subunit_trace will print all
        text attachments on successful test execution.
    :param bool show_binary_attachments: When set to true, subunit_trace will
        print binary attachments in addition to text attachments.
    :param str pdb: Takes in a single test_id to bypasses test
        discover and just execute the test specified without launching any
        additional processes. A file name may be used in place of a test name.

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    try:
        repo = util.get_repo_open(repo_url=repo_url)
    # If a repo is not found, and there a stestr config exists just create it
    except repository.RepositoryNotFound:
        conf = config_file.TestrConf.load_from_file(config)
        if not conf.test_path and not test_path:
            msg = (
                "No config file found and --test-path not specified. "
                "Either create or specify a .stestr.conf, tox.ini, "
                "or pyproject.toml, or use --test-path"
            )
            stdout.write(msg)
            exit(1)
        try:
            repo = util.get_repo_initialise(repo_url=repo_url)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            repo_path = repo_url or "./stestr"
            stdout.write(
                "The specified repository directory %s already "
                "exists. Please check if the repository already "
                "exists or select a different path\n" % repo_path
            )
            return 1

    combine_id = None
    concurrency = _to_int(concurrency)

    if concurrency and concurrency < 0:
        msg = (
            "The provided concurrency value: %s is not valid. An integer "
            ">= 0 must be used.\n" % concurrency
        )
        stdout.write(msg)
        return 2
    if combine:
        latest_id = repo.latest_id()
        combine_id = str(latest_id)
    if no_discover and pdb:
        msg = (
            "--no-discover and --pdb are mutually exclusive options, "
            "only specify one at a time"
        )
        stdout.write(msg)
        return 2
    if pdb and until_failure:
        msg = (
            "pdb mode does not function with the --until-failure flag, "
            "only specify one at a time"
        )
        stdout.write(msg)
        return 2

    if no_discover:
        ids = no_discover
        klass = None
        if "::" in ids:
            ids, klass = ids.split("::", 1)
            klass = ".".join(klass.split("::"))
        if ids.find(os.path.sep) != -1:
            root, _ = os.path.splitext(os.path.normpath(ids))
            ids = ".".join(root.split(os.path.sep))
        if klass:
            ids = f"{ids}.{klass}"
        stestr_python = sys.executable
        if os.environ.get("PYTHON"):
            python_bin = os.environ.get("PYTHON")
        elif stestr_python:
            python_bin = stestr_python
        else:
            raise RuntimeError(
                "The Python interpreter was not found and " "PYTHON is not set"
            )
        run_cmd = python_bin + " -m stestr.subunit_runner.run " + ids

        def run_tests():
            run_proc = [
                (
                    "subunit",
                    output.ReturnCodeToSubunit(
                        subprocess.Popen(run_cmd, shell=True, stdout=subprocess.PIPE)
                    ),
                )
            ]
            return load.load(
                in_streams=run_proc,
                subunit_out=subunit_out,
                repo_url=repo_url,
                run_id=combine_id,
                pretty_out=pretty_out,
                color=color,
                stdout=stdout,
                abbreviate=abbreviate,
                suppress_attachments=suppress_attachments,
                all_attachments=all_attachments,
                show_binary_attachments=show_binary_attachments,
            )

        if not until_failure:
            return run_tests()
        else:
            while True:
                result = run_tests()
                # If we're using subunit output we want to make sure to check
                # the result from the repository because load() returns 0
                # always on subunit output
                if subunit:
                    summary = testtools.StreamSummary()
                    last_run = repo.get_latest_run().get_subunit_stream()
                    stream = subunit.ByteStreamToStreamResult(last_run)
                    summary.startTestRun()
                    try:
                        stream.run(summary)
                    finally:
                        summary.stopTestRun()
                    if not results.wasSuccessful(summary):
                        result = 1
                if result:
                    return result

    if pdb:
        ids = pdb
        klass = None
        if "::" in ids:
            ids, klass = ids.split("::", 1)
            klass = ".".join(klass.split("::"))
        if ids.find(os.path.sep) != -1:
            root, _ = os.path.splitext(os.path.normpath(ids))
            ids = ".".join(root.split(os.path.sep))
        if klass:
            ids = f"{ids}.{klass}"
        runner = subunit_run.SubunitTestRunner
        stream = io.BytesIO()
        program.TestProgram(
            module=None,
            argv=["stestr", ids],
            testRunner=functools.partial(runner, stdout=stream),
        )
        stream.seek(0)
        run_proc = [("subunit", stream)]
        return load.load(
            in_streams=run_proc,
            subunit_out=subunit_out,
            repo_url=repo_url,
            run_id=combine_id,
            pretty_out=pretty_out,
            color=color,
            stdout=stdout,
            abbreviate=abbreviate,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=show_binary_attachments,
        )

    if failing or analyze_isolation:
        ids = _find_failing(repo)
    else:
        ids = None
    if load_list:
        list_ids = set()
        # Should perhaps be text.. currently does its own decode.
        with open(load_list, "rb") as list_file:
            list_ids = set(parse_list(list_file.read()))
        if ids is None:
            # Use the supplied list verbatim
            ids = list_ids
        else:
            # We have some already limited set of ids, just reduce to ids
            # that are both failing and listed.
            ids = list_ids.intersection(ids)

    conf = config_file.TestrConf.load_from_file(config)
    if not analyze_isolation:
        cmd = conf.get_run_command(
            ids,
            regexes=filters,
            group_regex=group_regex,
            repo_url=repo_url,
            serial=serial,
            worker_path=worker_path,
            concurrency=concurrency,
            exclude_list=exclude_list,
            include_list=include_list,
            exclude_regex=exclude_regex,
            top_dir=top_dir,
            test_path=test_path,
            randomize=random,
        )
        if isolated:
            result = 0
            cmd.setUp()
            try:
                ids = cmd.list_tests()
            finally:
                cmd.cleanUp()
            for test_id in ids:
                # TODO(mtreinish): add regex
                cmd = conf.get_run_command(
                    [test_id],
                    filters,
                    group_regex=group_regex,
                    repo_url=repo_url,
                    serial=serial,
                    worker_path=worker_path,
                    concurrency=concurrency,
                    exclude_list=exclude_list,
                    include_list=include_list,
                    exclude_regex=exclude_regex,
                    randomize=random,
                    test_path=test_path,
                    top_dir=top_dir,
                )

                run_result = _run_tests(
                    cmd,
                    until_failure,
                    subunit_out=subunit_out,
                    combine_id=combine_id,
                    repo_url=repo_url,
                    pretty_out=pretty_out,
                    color=color,
                    abbreviate=abbreviate,
                    stdout=stdout,
                    suppress_attachments=suppress_attachments,
                    all_attachments=all_attachments,
                    show_binary_attachments=show_binary_attachments,
                )
                if run_result > result:
                    result = run_result
            return result
        else:
            return _run_tests(
                cmd,
                until_failure,
                subunit_out=subunit_out,
                combine_id=combine_id,
                repo_url=repo_url,
                pretty_out=pretty_out,
                color=color,
                stdout=stdout,
                abbreviate=abbreviate,
                suppress_attachments=suppress_attachments,
                all_attachments=all_attachments,
                show_binary_attachments=show_binary_attachments,
            )
    else:
        # Where do we source data about the cause of conflicts.
        latest_run = repo.get_latest_run()
        # Stage one: reduce the list of failing tests (possibly further
        # reduced by testfilters) to eliminate fails-on-own tests.
        spurious_failures = set()
        for test_id in ids:
            # TODO(mtrienish): Add regex
            cmd = conf.get_run_command(
                [test_id],
                group_regex=group_regex,
                repo_url=repo_url,
                serial=serial,
                worker_path=worker_path,
                concurrency=concurrency,
                exclude_list=exclude_list,
                include_list=include_list,
                exclude_regex=exclude_regex,
                randomize=random,
                test_path=test_path,
                top_dir=top_dir,
            )
            if not _run_tests(cmd, until_failure):
                # If the test was filtered, it won't have been run.
                if test_id in repo.get_test_ids(repo.latest_id()):
                    spurious_failures.add(test_id)
                # This is arguably ugly, why not just tell the system that
                # a pass here isn't a real pass? [so that when we find a
                # test that is spuriously failing, we don't forget
                # that it is actually failing.
                # Alternatively, perhaps this is a case for data mining:
                # when a test starts passing, keep a journal, and allow
                # digging back in time to see that it was a failure,
                # what it failed with etc...
                # The current solution is to just let it get marked as
                # a pass temporarily.
        if not spurious_failures:
            # All done.
            return 0
        bisect_runner = bisect_tests.IsolationAnalyzer(
            latest_run,
            conf,
            _run_tests,
            repo,
            test_path=test_path,
            top_dir=top_dir,
            group_regex=group_regex,
            repo_url=repo_url,
            serial=serial,
            concurrency=concurrency,
        )
        # spurious-failure -> cause.
        return bisect_runner.bisect_tests(spurious_failures)


def _run_tests(
    cmd,
    until_failure,
    subunit_out=False,
    combine_id=None,
    repo_url=None,
    pretty_out=True,
    color=False,
    stdout=sys.stdout,
    abbreviate=False,
    suppress_attachments=False,
    all_attachments=False,
    show_binary_attachments=False,
):
    """Run the tests cmd was parameterised with."""
    cmd.setUp()
    try:

        def run_tests():
            run_procs = [
                ("subunit", output.ReturnCodeToSubunit(proc))
                for proc in cmd.run_tests()
            ]
            if not run_procs:
                stdout.write("The specified regex doesn't match with anything")
                return 1
            return load.load(
                (None, None),
                in_streams=run_procs,
                subunit_out=subunit_out,
                repo_url=repo_url,
                run_id=combine_id,
                pretty_out=pretty_out,
                color=color,
                stdout=stdout,
                abbreviate=abbreviate,
                suppress_attachments=suppress_attachments,
                all_attachments=all_attachments,
                show_binary_attachments=show_binary_attachments,
            )

        if not until_failure:
            return run_tests()
        else:
            while True:
                result = run_tests()
                # If we're using subunit output we want to make sure to check
                # the result from the repository because load() returns 0
                # always on subunit output
                if subunit_out:
                    repo = util.get_repo_open(repo_url=repo_url)
                    summary = testtools.StreamSummary()
                    last_run = repo.get_latest_run().get_subunit_stream()
                    stream = subunit.ByteStreamToStreamResult(last_run)
                    summary.startTestRun()
                    try:
                        stream.run(summary)
                    finally:
                        summary.stopTestRun()
                    if not results.wasSuccessful(summary):
                        result = 1
                if result:
                    return result
    finally:
        cmd.cleanUp()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import random
from unittest import mock

import ddt
import fixtures

from stestr import config_file
from stestr.tests import base


@ddt.ddt
class TestTestrConf(base.TestCase):
    @mock.patch.object(config_file.configparser, "ConfigParser")
    def setUp(self, mock_ConfigParser):
        super().setUp()
        self._testr_conf = config_file.TestrConf(mock.sentinel.config_file)

    @mock.patch.object(config_file.util, "get_repo_open")
    @mock.patch.object(config_file.test_processor, "TestProcessorFixture")
    @mock.patch.object(config_file, "sys")
    @mock.patch("os.path.exists", new=lambda x: True)
    def _check_get_run_command(
        self,
        mock_sys,
        mock_TestProcessorFixture,
        mock_get_repo_open,
        platform="win32",
        group_regex=".*",
        parallel_class=False,
        sys_executable="/usr/bin/python",
        expected_python="/usr/bin/python",
        expected_group_callback=mock.ANY,
        environment=None,
    ):
        mock_sys.platform = platform
        mock_sys.executable = sys_executable
        if environment is None:
            environment = {"PYTHON": ""}

        with mock.patch.dict("os.environ", environment):
            fixture = self._testr_conf.get_run_command(
                test_path="fake_test_path",
                top_dir="fake_top_dir",
                group_regex=group_regex,
                parallel_class=parallel_class,
            )

        self.assertEqual(mock_TestProcessorFixture.return_value, fixture)
        mock_get_repo_open.assert_called_once_with(repo_url=None)
        command = (
            '"%s" -m stestr.subunit_runner.run discover -t "%s" "%s" '
            "$LISTOPT $IDOPTION" % (expected_python, "fake_top_dir", "fake_test_path")
        )
        # Ensure TestProcessorFixture is created with defaults except for where
        # we specfied and with the correct python.
        mock_TestProcessorFixture.assert_called_once_with(
            None,
            command,
            "--list",
            "--load-list $IDFILE",
            mock_get_repo_open.return_value,
            exclude_regex=None,
            exclude_list=None,
            concurrency=0,
            group_callback=expected_group_callback,
            test_filters=None,
            randomize=False,
            serial=False,
            include_list=None,
            worker_path=None,
        )

    @mock.patch.object(config_file, "sys")
    def _check_get_run_command_exception(
        self,
        mock_sys,
        platform="win32",
        sys_executable="/usr/bin/python",
        environment=None,
    ):
        mock_sys.platform = platform
        mock_sys.executable = sys_executable
        if environment is None:
            environment = {"PYTHON": ""}

        with mock.patch.dict("os.environ", environment):
            self.assertRaises(
                RuntimeError,
                self._testr_conf.get_run_command,
                test_path="fake_test_path",
                top_dir="fake_top_dir",
            )

    def test_get_run_command_linux(self):
        self._check_get_run_command(
            platform="linux2", expected_python="/usr/bin/python"
        )

    def test_get_run_command_emptysysexecutable_noenv(self):
        self._check_get_run_command_exception(platform="linux2", sys_executable=None)

    def test_get_run_command_emptysysexecutable_win32(self):
        self._check_get_run_command_exception(
            platform="win32", sys_executable=None, environment={"PYTHON": "python3"}
        )

    def test_get_run_command_emptysysexecutable_withenv(self):
        self._check_get_run_command(
            platform="linux2",
            sys_executable=None,
            expected_python="${PYTHON}",
            environment={"PYTHON": "/usr/bin/python3"},
        )

    def test_get_run_command_win32(self):
        self._check_get_run_command()

    def test_get_run_command_parallel_class(self):
        self._check_get_run_command(parallel_class=True)

    def test_get_run_command_nogroup_regex_noparallel_class(self):
        self._testr_conf.parallel_class = False
        self._testr_conf.group_regex = ""
        self._check_get_run_command(group_regex="", expected_group_callback=None)

    @ddt.data((".\\", ".\\\\"), ("a\\b\\", "a\\b\\\\"), ("a\\b", "a\\b"))
    @ddt.unpack
    @mock.patch("os.sep", new="\\")
    def test_sanitize_dir_win32(self, path, expected):
        sanitized = self._testr_conf._sanitize_path(path)
        self.assertEqual(expected, sanitized)

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.config_file.TestrConf.__init__")
    def test_load_from_file_user_specified(self, initializer, isfile):
        # Test that user-specified config files are "one-and-done"
        initializer.return_value = None
        isfile.return_value = True
        config_file.TestrConf.load_from_file("user.conf")
        initializer.assert_called_once_with("user.conf")

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.config_file.TestrConf.__init__")
    def test_load_from_file_user_specified_fails(self, initializer, isfile):
        # Test that user-specified config files that do not exist gives up
        # immediately
        initializer.return_value = None
        initializer.side_effect = FileNotFoundError
        isfile.return_value = False
        self.assertRaises(
            FileNotFoundError, config_file.TestrConf.load_from_file, "user.conf"
        )
        isfile.assert_called_once_with("user.conf")
        initializer.assert_called_once_with("user.conf")

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.config_file.TestrConf.__init__")
    def test_load_from_file_toml_has_precedence(self, initializer, isfile):
        # Test that tox.ini is ignored if a pyproject.toml config exists
        initializer.return_value = None
        isfile.return_value = False
        config_file.TestrConf.load_from_file(".stestr.conf")
        isfile.assert_called_once_with(".stestr.conf")
        initializer.assert_called_once_with("pyproject.toml")

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.config_file.TestrConf.__init__")
    def test_load_from_file_ini_fallback(self, initializer, isfile):
        initializer.return_value = None
        # The only difference between "no config file" and "nothing defined
        # in the config file" is the type of exception thrown; we'll make
        # sure that, in aggregate, we test for both conditions
        exc = random.choice([FileNotFoundError, KeyError])
        initializer.side_effect = (exc, None)
        isfile.return_value = False
        config_file.TestrConf.load_from_file(".stestr.conf")
        isfile.assert_called_once_with(".stestr.conf")
        initializer.assert_has_calls(
            [mock.call("pyproject.toml"), mock.call("tox.ini", section="stestr")]
        )

    @mock.patch.object(config_file.tomlkit, "load")
    def test_toml_load(self, mock_toml):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        file_path = os.path.join(tmpdir, "myfile.toml")
        with open(file_path, "w"):
            pass
        self._testr_conf = config_file.TestrConf(file_path)
        self._check_get_run_command()
        mock_toml.return_value.__getitem__.assert_called_once_with("tool")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from cliff import app
from cliff import commandmanager
from stestr import __version__


class StestrCLI(app.App):
    def __init__(self):
        super().__init__(
            description="A parallel Python test runner built around subunit",
            version=__version__,
            command_manager=commandmanager.CommandManager("stestr.cm"),
            deferred_help=True,
        )

    def initialize_app(self, argv):
        self.options.debug = True
        self.LOG.debug("initialize_app")

    def prepare_to_run_command(self, cmd):
        self.LOG.debug("prepare_to_run_command %s", cmd.__class__.__name__)
        group_regex = (
            r"([^\.]*\.)*" if cmd.app_args.parallel_class else cmd.app_args.group_regex
        )
        cmd.app_args.group_regex = group_regex

    def clean_up(self, cmd, result, err):
        self.LOG.debug("clean_up %s", cmd.__class__.__name__)
        if err:
            self.LOG.debug("got an error: %s", err)

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super().build_option_parser(description, version, argparse_kwargs)
        parser = self._set_common_opts(parser)
        return parser

    def _set_common_opts(self, parser):
        parser.add_argument(
            "--user-config",
            dest="user_config",
            default=None,
            help="An optional path to a default user config "
            "file if one is not specified ~/.stestr.yaml "
            "and ~/.config/stestr.yaml will be tried in "
            "that order",
        )
        parser.add_argument(
            "-d",
            "--here",
            dest="here",
            help="Set the directory or url that a command "
            "should run from. This affects all default "
            "path lookups but does not affect paths "
            "supplied to the command.",
            default=None,
            type=str,
        )
        parser.add_argument(
            "--config",
            "-c",
            dest="config",
            default=".stestr.conf",
            help="Set a stestr config file to use with this "
            "command. If one isn't specified then "
            ".stestr.conf in the directory that a command"
            " is running from is used",
        )
        parser.add_argument(
            "--repo-url",
            "-u",
            dest="repo_url",
            default=None,
            help="Set the repo url to use. This should just "
            "be the path to the stestr repository "
            "directory",
        )
        parser.add_argument(
            "--test-path",
            "-t",
            dest="test_path",
            default=None,
            help="Set the test path to use for unittest "
            "discovery. If both this and the "
            "corresponding config file option are set, "
            "this value will be used.",
        )
        parser.add_argument(
            "--top-dir",
            dest="top_dir",
            default=None,
            help="Set the top dir to use for unittest "
            "discovery. If both this and the "
            "corresponding config file option are set, "
            "this value will be used.",
        )
        parser.add_argument(
            "--group-regex",
            "--group_regex",
            "-g",
            dest="group_regex",
            default=None,
            help="Set a group regex to use for grouping tests"
            " together in the stestr scheduler. If "
            "both this and the corresponding config file "
            "option are set this value will be used.",
        )
        parser.add_argument(
            "--parallel-class",
            "-p",
            action="store_true",
            default=False,
            help="Set the flag to group tests by class. NOTE: "
            "This flag takes priority over the "
            "`--group-regex` option even if it's set.",
        )

        return parser


def main(argv=sys.argv[1:]):
    cli = StestrCLI()
    return cli.run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import re
from unittest import mock

from subunit import iso8601

from stestr.repository import memory
from stestr import scheduler
from stestr.tests import base


class TestScheduler(base.TestCase):
    def _add_timed_test(self, id, duration, result):
        start = datetime.datetime.now()
        start = start.replace(tzinfo=iso8601.UTC)
        result.status(test_id=id, test_status="inprogress", timestamp=start)
        timestamp = start + datetime.timedelta(seconds=duration)
        result.status(test_id=id, test_status="success", timestamp=timestamp)

    def test_partition_tests(self):
        repo = memory.RepositoryFactory().initialise("memory:")
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("slow", 3, result)
        self._add_timed_test("fast1", 1, result)
        self._add_timed_test("fast2", 1, result)
        result.stopTestRun()
        test_ids = frozenset(
            ["slow", "fast1", "fast2", "unknown1", "unknown2", "unknown3", "unknown4"]
        )
        partitions = scheduler.partition_tests(test_ids, 2, repo, None)
        self.assertTrue("slow" in partitions[0])
        self.assertFalse("fast1" in partitions[0])
        self.assertFalse("fast2" in partitions[0])
        self.assertFalse("slow" in partitions[1])
        self.assertTrue("fast1" in partitions[1])
        self.assertTrue("fast2" in partitions[1])
        self.assertEqual(3, len(partitions[0]))
        self.assertEqual(4, len(partitions[1]))

    def test_random_partitions(self):
        repo = memory.RepositoryFactory().initialise("memory:")
        test_ids = frozenset(["a_test", "b_test", "c_test", "d_test"])
        random_parts = scheduler.partition_tests(
            test_ids, 2, repo, None, randomize=True
        )
        # NOTE(masayukig): We can't test this randomness. So just checking
        # what we should get here.
        self.assertEqual(2, len(random_parts))
        self.assertTrue(isinstance(random_parts, list))
        self.assertTrue(isinstance(random_parts[0], list))
        self.assertTrue(isinstance(random_parts[1], list))
        flatten_random_parts = []
        for i, j in random_parts:
            flatten_random_parts.append(i)
            flatten_random_parts.append(j)
        for i in test_ids:
            self.assertIn(i, flatten_random_parts)

    def test_partition_tests_with_zero_duration(self):
        repo = memory.RepositoryFactory().initialise("memory:")
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("zero1", 0, result)
        self._add_timed_test("zero2", 0, result)
        result.stopTestRun()
        # Partitioning by two should generate two one-entry partitions.
        test_ids = frozenset(["zero1", "zero2"])
        partitions = scheduler.partition_tests(test_ids, 2, repo, None)
        self.assertEqual(1, len(partitions[0]))
        self.assertEqual(1, len(partitions[1]))

    def test_partition_tests_with_grouping(self):
        repo = memory.RepositoryFactory().initialise("memory:")
        result = repo.get_inserter()
        result.startTestRun()
        self._add_timed_test("TestCase1.slow", 3, result)
        self._add_timed_test("TestCase2.fast1", 1, result)
        self._add_timed_test("TestCase2.fast2", 1, result)
        result.stopTestRun()
        test_ids = frozenset(
            [
                "TestCase1.slow",
                "TestCase1.fast",
                "TestCase1.fast2",
                "TestCase2.fast1",
                "TestCase3.test1",
                "TestCase3.test2",
                "TestCase2.fast2",
                "TestCase4.test",
                "testdir.testfile.TestCase5.test",
            ]
        )

        def group_id(test_id, regex=re.compile("TestCase[0-5]")):
            match = regex.match(test_id)
            if match:
                return match.group(0)

        partitions = scheduler.partition_tests(test_ids, 2, repo, group_id)
        # Timed groups are deterministic:
        self.assertTrue("TestCase2.fast1" in partitions[0])
        self.assertTrue("TestCase2.fast2" in partitions[0])
        self.assertTrue("TestCase1.slow" in partitions[1])
        self.assertTrue("TestCase1.fast" in partitions[1])
        self.assertTrue("TestCase1.fast2" in partitions[1])
        # Untimed groups just need to be in the same partition:
        if "TestCase3.test1" in partitions[0]:
            self.assertTrue("TestCase3.test2" in partitions[0])
        if "TestCase4.test" not in partitions[0]:
            self.assertTrue("TestCase4.test" in partitions[1])
        if "testdir.testfile.TestCase5.test" not in partitions[0]:
            self.assertTrue("testdir.testfile.TestCase5.test" in partitions[1])

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions(self):
        test_ids = ["test_a", "test_b", "your_test"]
        fake_worker_yaml = [
            {"worker": ["test_"]},
            {"worker": ["test"]},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, "fakepath")
        expected_grouping = [
            ["test_a", "test_b"],
            ["test_a", "test_b", "your_test"],
        ]
        self.assertEqual(expected_grouping, groups)

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions_group_without_list(self):
        test_ids = ["test_a", "test_b", "your_test"]
        fake_worker_yaml = [
            {"worker": ["test_"]},
            {"worker": "test"},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            self.assertRaises(
                TypeError, scheduler.generate_worker_partitions, test_ids, "fakepath"
            )

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions_no_worker_tag(self):
        test_ids = ["test_a", "test_b", "your_test"]
        fake_worker_yaml = [
            {"worker-foo": ["test_"]},
            {"worker": ["test"]},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            self.assertRaises(
                TypeError, scheduler.generate_worker_partitions, test_ids, "fakepath"
            )

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions_group_without_match(self):
        test_ids = ["test_a", "test_b", "your_test"]
        fake_worker_yaml = [
            {"worker": ["test_"]},
            {"worker": ["test"]},
            {"worker": ["foo"]},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, "fakepath")
        expected_grouping = [
            ["test_a", "test_b"],
            ["test_a", "test_b", "your_test"],
        ]
        self.assertEqual(expected_grouping, groups)

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions_with_count(self):
        test_ids = ["test_a", "test_b", "your_test", "a_thing1", "a_thing2"]
        fake_worker_yaml = [
            {"worker": ["test_"]},
            {"worker": ["test"]},
            {"worker": ["a_thing"], "concurrency": 2},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, "fakepath")
        expected_grouping = [
            ["test_a", "test_b"],
            ["test_a", "test_b", "your_test"],
            ["a_thing1"],
            ["a_thing2"],
        ]
        for worker in expected_grouping:
            self.assertIn(worker, groups)

    @mock.patch("builtins.open", mock.mock_open(), create=True)
    def test_generate_worker_partitions_with_count_1(self):
        test_ids = ["test_a", "test_b", "your_test"]
        fake_worker_yaml = [
            {"worker": ["test_"]},
            {"worker": ["test"], "count": 1},
        ]
        with mock.patch("yaml.safe_load", return_value=fake_worker_yaml):
            groups = scheduler.generate_worker_partitions(test_ids, "fakepath")
        expected_grouping = [
            ["test_a", "test_b"],
            ["test_a", "test_b", "your_test"],
        ]
        self.assertEqual(expected_grouping, groups)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil
import tempfile
from unittest import mock

from stestr.repository import util
from stestr.tests import base


class TestUtil(base.TestCase):
    def setUp(self):
        super().setUp()
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)
        cwd = os.getcwd()
        os.chdir(self.temp_dir)
        self.temp_dir = os.getcwd()
        self.addCleanup(os.chdir, cwd)

    def test_get_default_url_file(self):
        repo_url = util._get_default_repo_url("file")
        self.assertEqual(self.temp_dir, repo_url)

    def test_get_default_url_invalid_type(self):
        self.assertRaises(TypeError, util._get_default_repo_url, "invalid_type")

    @mock.patch("importlib.import_module", side_effect=ImportError)
    def test_non_sql_get_repo_init_no_deps_import_error(self, import_mock):
        self.assertRaises(ImportError, util.get_repo_initialise, "file")

    @mock.patch("importlib.import_module", side_effect=ImportError)
    def test_non_sql_get_repo_open_no_deps_import_error(self, import_mock):
        self.assertRaises(ImportError, util.get_repo_open, "file")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io

from stestr.commands import load
from stestr.tests import base


class TestLoadCommand(base.TestCase):
    def test_empty_with_pretty_out(self):
        stream = io.BytesIO()
        output = io.BytesIO()
        res = load.load(
            in_streams=[("subunit", stream)], pretty_out=True, stdout=output
        )
        self.assertEqual(1, res)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Persistent storage of test results."""

import errno
from io import BytesIO
from operator import methodcaller
import os
import sys
import tempfile

from dbm import dumb as my_dbm
from subunit import TestProtocolClient
import subunit.v2
import testtools

from stestr.repository import abstract as repository
from stestr import utils


def atomicish_rename(source, target):
    if os.name != "posix" and os.path.exists(target):
        os.remove(target)
    os.rename(source, target)


class RepositoryFactory(repository.AbstractRepositoryFactory):
    def initialise(klass, url):
        """Create a repository at url/path."""
        base = os.path.join(os.path.expanduser(url), ".stestr")
        try:
            os.mkdir(base)
        except OSError as e:
            if e.errno == errno.EEXIST and not os.listdir(base):
                # It shouldn't be harmful initializing an empty dir
                pass
            else:
                raise
        with open(os.path.join(base, "format"), "wt") as stream:
            stream.write("1\n")
        result = Repository(base)
        result._write_next_stream(0)
        return result

    def open(self, url):
        path = os.path.expanduser(url)
        base = os.path.join(path, ".stestr")
        try:
            stream = open(os.path.join(base, "format"))
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise repository.RepositoryNotFound(url)
            raise
        with stream:
            if "1\n" != stream.read():
                raise ValueError(url)
        return Repository(base)


class Repository(repository.AbstractRepository):
    """Disk based storage of test results.

    This repository stores each stream it receives as a file in a directory.
    Indices are then built on top of this basic store.

    This particular disk layout is subject to change at any time, as its
    primarily a bootstrapping exercise at this point. Any changes made are
    likely to have an automatic upgrade process.
    """

    def __init__(self, base):
        """Create a file-based repository object for the repo at 'base'.

        :param base: The path to the repository.
        """
        self.base = base

    def _allocate(self):
        # XXX: lock the file. K?!
        value = self.count()
        self._write_next_stream(value + 1)
        return value

    def _next_stream(self):
        with open(os.path.join(self.base, "next-stream")) as fp:
            next_content = fp.read()
        try:
            return int(next_content)
        except ValueError:
            raise ValueError("Corrupt next-stream file: %r" % next_content)

    def count(self):
        return self._next_stream()

    def latest_id(self):
        result = self._next_stream() - 1
        if result < 0:
            raise KeyError("No tests in repository")
        return result

    def get_run_ids(self):
        result = []
        max_id = self._next_stream()
        if max_id < 0:
            raise KeyError("No tests in repository")
        for i in range(max_id):
            if os.path.isfile(os.path.join(self.base, str(i))):
                result.append(str(i))
        return result

    def remove_run_id(self, run_id):
        run_path = os.path.join(self.base, run_id)
        if not os.path.isfile(run_path):
            raise KeyError("No run %s in repository" % run_id)
        os.remove(run_path)

    def get_failing(self):
        try:
            with open(os.path.join(self.base, "failing"), "rb") as fp:
                run_subunit_content = fp.read()
        except OSError:
            err = sys.exc_info()[1]
            if err.errno == errno.ENOENT:
                run_subunit_content = b""
            else:
                raise
        return _DiskRun(None, run_subunit_content)

    def _get_metadata(self, run_id):
        db = my_dbm.open(self._path("meta.dbm"), "c")
        try:
            metadata = db.get(str(run_id))
        finally:
            db.close()
        return metadata

    def get_test_run(self, run_id):
        try:
            with open(os.path.join(self.base, str(run_id)), "rb") as fp:
                run_subunit_content = fp.read()
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise KeyError("No such run.")
            else:
                raise
        metadata = self._get_metadata(run_id)
        return _DiskRun(run_id, run_subunit_content, metadata=metadata)

    def _get_inserter(self, partial, run_id=None, metadata=None):
        return _Inserter(self, partial, run_id, metadata=metadata)

    def _get_test_times(self, test_ids):
        # May be too slow, but build and iterate.
        # 'c' because an existing repo may be missing a file.
        try:
            db = my_dbm.open(self._path("times.dbm"), "c")
        except my_dbm.error:
            os.remove(self._path("times.dbm"))
            db = my_dbm.open(self._path("times.dbm"), "c")
        try:
            result = {}
            for test_id in test_ids:
                if type(test_id) != str:
                    test_id = test_id.encode("utf8")
                stripped_test_id = utils.cleanup_test_name(test_id)
                # gdbm does not support get().
                try:
                    duration = db[stripped_test_id]
                except KeyError:
                    duration = None
                if duration is not None:
                    result[test_id] = float(duration)
            return result
        finally:
            db.close()

    def _path(self, suffix):
        return os.path.join(self.base, suffix)

    def _write_next_stream(self, value):
        # Note that this is unlocked and not threadsafe : single
        # user, repo-per-working-tree model makes this acceptable in the short
        # term. Likewise we don't fsync - this data isn't valuable enough to
        # force disk IO.
        prefix = self._path("next-stream")
        with open(prefix + ".new", "wt") as stream:
            stream.write("%d\n" % value)
        atomicish_rename(prefix + ".new", prefix)

    def find_metadata(self, metadata):
        run_ids = []
        db = my_dbm.open(self._path("meta.dbm"), "c")
        try:
            for run_id in db:
                if db.get(run_id) == metadata:
                    run_ids.append(run_id)
        finally:
            db.close()
        return run_ids


class _DiskRun(repository.AbstractTestRun):
    """A test run that was inserted into the repository."""

    def __init__(self, run_id, subunit_content, metadata=None):
        """Create a _DiskRun with the content subunit_content."""
        self._run_id = run_id
        self._content = subunit_content
        assert type(subunit_content) is bytes
        self._metadata = metadata

    def get_id(self):
        return self._run_id

    def get_subunit_stream(self):
        # Transcode - we want V2.
        v1_stream = BytesIO(self._content)
        v1_case = subunit.ProtocolTestCase(v1_stream)
        output = BytesIO()
        output_stream = subunit.v2.StreamResultToBytes(output)
        output_stream = testtools.ExtendedToStreamDecorator(output_stream)
        output_stream.startTestRun()
        try:
            v1_case.run(output_stream)
        finally:
            output_stream.stopTestRun()
        output.seek(0)
        return output

    def get_test(self):
        # case = subunit.ProtocolTestCase(self.get_subunit_stream())
        case = subunit.ProtocolTestCase(BytesIO(self._content))

        def wrap_result(result):
            # Wrap in a router to mask out startTestRun/stopTestRun from the
            # ExtendedToStreamDecorator.
            result = testtools.StreamResultRouter(result, do_start_stop_run=False)
            # Wrap that in ExtendedToStreamDecorator to convert v1 calls to
            # StreamResult.
            return testtools.ExtendedToStreamDecorator(result)

        return testtools.DecorateTestCaseResult(
            case, wrap_result, methodcaller("startTestRun"), methodcaller("stopTestRun")
        )

    def get_metadata(self):
        return self._metadata


class _SafeInserter:
    def __init__(self, repository, partial=False, run_id=None, metadata=None):
        # XXX: Perhaps should factor into a decorator and use an unaltered
        # TestProtocolClient.
        self._repository = repository
        self._run_id = run_id
        self._metadata = metadata
        if not self._run_id:
            fd, name = tempfile.mkstemp(dir=self._repository.base)
            self.fname = name
            stream = os.fdopen(fd, "wb")
        else:
            self.fname = os.path.join(self._repository.base, self._run_id)
            stream = open(self.fname, "ab")
        self.partial = partial
        # The time take by each test, flushed at the end.
        self._times = {}
        self._test_start = None
        self._time = None
        subunit_client = testtools.StreamToExtendedDecorator(TestProtocolClient(stream))
        self.hook = testtools.CopyStreamResult(
            [subunit_client, testtools.StreamToDict(self._handle_test)]
        )
        self._stream = stream

    def _handle_test(self, test_dict):
        start, stop = test_dict["timestamps"]
        if test_dict["status"] == "exists" or None in (start, stop):
            return
        test_id = utils.cleanup_test_name(test_dict["id"])
        self._times[test_id] = str((stop - start).total_seconds())

    def startTestRun(self):
        self.hook.startTestRun()

    def stopTestRun(self):
        self.hook.stopTestRun()
        self._stream.flush()
        self._stream.close()
        run_id = self._name()
        if not self._run_id:
            final_path = os.path.join(self._repository.base, str(run_id))
            atomicish_rename(self.fname, final_path)
        if self._metadata:
            db = my_dbm.open(self._repository._path("meta.dbm"), "c")
            try:
                dbm_run_id = str(run_id)
                db[dbm_run_id] = str(self._metadata)
            finally:
                db.close()

        # May be too slow, but build and iterate.
        db = my_dbm.open(self._repository._path("times.dbm"), "c")
        try:
            db_times = {}
            for key, value in self._times.items():
                if type(key) != str:
                    key = key.encode("utf8")
                db_times[key] = value
            if getattr(db, "update", None):
                db.update(db_times)
            else:
                for key, value in db_times.items():
                    db[key] = value
        finally:
            db.close()
        if not self._run_id:
            self._run_id = run_id

    def status(self, *args, **kwargs):
        self.hook.status(*args, **kwargs)

    def _cancel(self):
        """Cancel an insertion."""
        self._stream.close()
        os.unlink(self.fname)

    def get_id(self):
        return self._run_id


class _FailingInserter(_SafeInserter):
    """Insert a stream into the 'failing' file."""

    def _name(self):
        return "failing"


class _Inserter(_SafeInserter):
    def _name(self):
        if not self._run_id:
            return self._repository._allocate()
        else:
            return self._run_id

    def stopTestRun(self):
        super().stopTestRun()
        # XXX: locking (other inserts may happen while we update the failing
        # file).
        # Combine failing + this run : strip passed tests, add failures.
        # use memory repo to aggregate. a bit awkward on layering ;).
        # Should just pull the failing items aside as they happen perhaps.
        # Or use a router and avoid using a memory object at all.
        from stestr.repository import memory

        repo = memory.Repository()
        if self.partial:
            # Seed with current failing
            inserter = testtools.ExtendedToStreamDecorator(repo.get_inserter())
            inserter.startTestRun()
            failing = self._repository.get_failing()
            failing.get_test().run(inserter)
            inserter.stopTestRun()
        inserter = testtools.ExtendedToStreamDecorator(repo.get_inserter(partial=True))
        inserter.startTestRun()
        run = self._repository.get_test_run(self.get_id())
        run.get_test().run(inserter)
        inserter.stopTestRun()
        # and now write to failing
        inserter = _FailingInserter(self._repository)
        _inserter = testtools.ExtendedToStreamDecorator(inserter)
        _inserter.startTestRun()
        try:
            repo.get_failing().get_test().run(_inserter)
        except Exception:
            inserter._cancel()
            raise
        else:
            _inserter.stopTestRun()
        return self.get_id()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import importlib
import os
import warnings


def _get_default_repo_url(repo_type):
    if repo_type == "file":
        repo_url = os.getcwd()
    else:
        raise TypeError("Unrecognized repository type %s" % repo_type)
    return repo_url


def get_repo_open(repo_type=None, repo_url=None):
    """Return an already initialized repo object given the parameters

    :param str repo_type: DEPRECATED - The repo module to use for the returned
        repo
    :param str repo_url: An optional repo url, if one is not specified the
        default $CWD/.stestr will be used.
    """
    if repo_type is not None:
        msg = (
            "WARNING: Specifying repository type is deprecated and will be "
            "removed in future release.\n"
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=3)
    else:
        repo_type = "file"
    repo_module = importlib.import_module("stestr.repository." + repo_type)
    if not repo_url:
        repo_url = _get_default_repo_url(repo_type)
    return repo_module.RepositoryFactory().open(repo_url)


def get_repo_initialise(repo_type=None, repo_url=None):
    """Return a newly initialized repo object given the parameters

    :param str repo_type: DEPRECATED - The repo module to use for the returned
        repo
    :param str repo_url: An optional repo url, if one is not specified the
        default $CWD/.stestr will be used.
    """
    if repo_type is not None:
        msg = (
            "WARNING: Specifying repository type is deprecated and will be "
            "removed in future release.\n"
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=3)
    else:
        repo_type = "file"
    repo_module = importlib.import_module("stestr.repository." + repo_type)
    if not repo_url:
        repo_url = _get_default_repo_url(repo_type)
    return repo_module.RepositoryFactory().initialise(repo_url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io

from stestr.commands import run
from stestr.tests import base


class TestRunCommand(base.TestCase):
    def test_to_int_positive_int(self):
        self.assertEqual(29, run._to_int(29))

    def test_to_int_positive_int_str(self):
        self.assertEqual(42, run._to_int("42"))

    def test_to_int_negative_int(self):
        self.assertEqual(-2, run._to_int(-2))

    def test_to_int_negative_int_str(self):
        self.assertEqual(-45, run._to_int("-45"))

    def test_to_int_invalid_str(self):
        fake_stderr = io.StringIO()
        out = run._to_int("I am not an int", out=fake_stderr)
        expected = 'Unable to convert "I am not an int" to an integer.  ' "Using 0.\n"
        self.assertEqual(fake_stderr.getvalue(), expected)
        self.assertEqual(0, out)

    def test_to_int_none(self):
        fake_stderr = io.StringIO()
        out = run._to_int(None, out=fake_stderr)
        expected = 'Unable to convert "None" to an integer.  ' "Using 0.\n"
        self.assertEqual(fake_stderr.getvalue(), expected)
        self.assertEqual(0, out)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Handling of lists of tests - common code to --load-list etc."""

import io

try:
    from subunit import ByteStreamToStreamResult as bytestream_to_streamresult
except ImportError:
    bytestream_to_streamresult = None

try:
    from testtools.testresult.doubles import StreamResult as stream_result
except ImportError:
    stream_result = None


def write_list(stream, test_ids):
    """Write test_ids out to stream.

    :param stream: A file-like object.
    :param test_ids: An iterable of test ids.
    """
    # May need utf8 explicitly?
    stream.write(bytes(("\n".join(list(test_ids) + [""])).encode("utf8")))


def parse_list(list_bytes):
    """Parse list_bytes into a list of test ids."""
    return _v1(list_bytes)


def parse_enumeration(enumeration_bytes):
    """Parse enumeration_bytes into a list of test_ids."""
    # If subunit v2 is available, use it.
    if bytestream_to_streamresult is not None:
        return _v2(enumeration_bytes)
    else:
        return _v1(enumeration_bytes)


def _v1(list_bytes):
    return [id.strip() for id in list_bytes.decode("utf8").split("\n") if id.strip()]


def _v2(list_bytes):
    parser = bytestream_to_streamresult(
        io.BytesIO(list_bytes), non_subunit_name="stdout"
    )
    result = stream_result()
    parser.run(result)
    return [event[1] for event in result._events if event[2] == "exists"]



# Copyright 2015 SUSE Linux GmbH
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime as dt
import io
import os
import sys
from unittest.mock import patch

from ddt import data
from ddt import ddt
from ddt import unpack
from iso8601 import UTC
from subunit.v2 import StreamResultToBytes

from stestr import subunit_trace
from stestr.tests import base


@ddt
class TestSubunitTrace(base.TestCase):
    def setUp(self):
        super().setUp()
        # NOTE(mtreinish): subunit-trace relies on a global to track results
        # with the expectation that it's run once per python interpreter
        # (like per stestr run or other command). Make sure to clear those on
        # each test to isolate the tests from each other.
        subunit_trace.RESULTS = {}
        subunit_trace.FAILS = []

    @data(
        (
            [dt(2015, 4, 17, 22, 23, 14, 111111), dt(2015, 4, 17, 22, 23, 14, 111111)],
            "0.000000s",
        ),
        (
            [dt(2015, 4, 17, 22, 23, 14, 111111), dt(2015, 4, 17, 22, 23, 15, 111111)],
            "1.000000s",
        ),
        ([dt(2015, 4, 17, 22, 23, 14, 111111), None], ""),
    )
    @unpack
    def test_get_durating(self, timestamps, expected_result):
        self.assertEqual(subunit_trace.get_duration(timestamps), expected_result)

    @data(
        (
            [dt(2015, 4, 17, 22, 23, 14, 111111), dt(2015, 4, 17, 22, 23, 14, 111111)],
            0.0,
        ),
        (
            [dt(2015, 4, 17, 22, 23, 14, 111111), dt(2015, 4, 17, 22, 23, 15, 111111)],
            1.0,
        ),
        ([dt(2015, 4, 17, 22, 23, 14, 111111), None], 0.0),
    )
    @unpack
    def test_run_time(self, timestamps, expected_result):
        patched_res = {0: [{"timestamps": timestamps}]}
        with patch.dict(subunit_trace.RESULTS, patched_res, clear=True):
            self.assertEqual(subunit_trace.run_time(), expected_result)

    def test_trace(self):
        regular_stream = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "sample_streams/successful.subunit",
        )
        bytes_ = io.BytesIO()
        with open(regular_stream, "rb") as stream:
            bytes_.write(bytes(stream.read()))
        bytes_.seek(0)
        stdin = io.TextIOWrapper(io.BufferedReader(bytes_))
        returncode = subunit_trace.trace(stdin, sys.stdout)
        self.assertEqual(0, returncode)

    def test_trace_with_all_skips(self):
        regular_stream = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "sample_streams/all_skips.subunit",
        )
        bytes_ = io.BytesIO()
        with open(regular_stream, "rb") as stream:
            bytes_.write(bytes(stream.read()))
        bytes_.seek(0)
        stdin = io.TextIOWrapper(io.BufferedReader(bytes_))
        returncode = subunit_trace.trace(stdin, sys.stdout)
        self.assertEqual(1, returncode)

    def test_trace_with_failures(self):
        regular_stream = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "sample_streams/failure.subunit"
        )
        bytes_ = io.BytesIO()
        with open(regular_stream, "rb") as stream:
            bytes_.write(bytes(stream.read()))
        bytes_.seek(0)
        stdin = io.TextIOWrapper(io.BufferedReader(bytes_))
        returncode = subunit_trace.trace(stdin, sys.stdout)
        self.assertEqual(1, returncode)

    def test_trace_with_stuck_inprogress(self):
        output = io.BytesIO()
        stream = StreamResultToBytes(output)
        stream.startTestRun()
        stream.status(
            test_id="test_passes", test_status="inprogress", timestamp=dt.now(UTC)
        )
        stream.status(
            test_id="test_segfault", test_status="inprogress", timestamp=dt.now(UTC)
        )
        stream.status(
            test_id="test_passes", test_status="success", timestamp=dt.now(UTC)
        )
        stream.stopTestRun()
        output.seek(0)
        # capture stderr for test
        stderr = io.StringIO()
        sys_err = sys.stderr
        sys.stderr = stderr

        def restore_stderr():
            sys.stderr = sys_err

        self.addCleanup(restore_stderr)
        stdin = io.TextIOWrapper(io.BufferedReader(output))
        returncode = subunit_trace.trace(stdin, sys.stdout)
        self.assertEqual(1, returncode)
        stderr.seek(0)
        expected = """
The following tests exited without returning a status
and likely segfaulted or crashed Python:

\t* test_segfault
"""
        self.assertEqual(stderr.read(), expected)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# Copyright 2014 Samsung Electronics
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Trace a subunit stream in reasonable detail and high accuracy."""
import argparse
import functools
import os
import re
import sys

import subunit
import testtools

from stestr import __version__
from stestr import colorizer
from stestr import results

# NOTE(mtreinish) on python3 anydbm was renamed dbm and the python2 dbm module
# was renamed to dbm.ndbm, this block takes that into account
try:
    import anydbm as dbm
except ImportError:
    import dbm

DAY_SECONDS = 60 * 60 * 24
FAILS = []
RESULTS = {}


def total_seconds(timedelta):
    # NOTE(mtreinish): This method is built-in to the timedelta class in
    # python >= 2.7 it is here to enable it's use on older versions
    return (
        (timedelta.days * DAY_SECONDS + timedelta.seconds) * 10**6
        + timedelta.microseconds
    ) / 10**6


def cleanup_test_name(name, strip_tags=True, strip_scenarios=False):
    """Clean up the test name for display.

    By default we strip out the tags in the test because they don't help us
    in identifying the test that is run to it's result.

    Make it possible to strip out the testscenarios information (not to
    be confused with tempest scenarios) however that's often needed to
    indentify generated negative tests.
    """
    if strip_tags:
        tags_start = name.find("[")
        tags_end = name.find("]")
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1 :]
            name = newname

    if strip_scenarios:
        tags_start = name.find("(")
        tags_end = name.find(")")
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1 :]
            name = newname

    return name


def get_duration(timestamps):
    start, end = timestamps
    if not start or not end:
        duration = ""
    else:
        delta = end - start
        duration = "%d.%06ds" % (
            delta.days * DAY_SECONDS + delta.seconds,
            delta.microseconds,
        )
    return duration


def find_worker(test):
    """Get the worker number.

    If there are no workers because we aren't in a concurrent environment,
    assume the worker number is 0.
    """
    for tag in test["tags"]:
        if tag.startswith("worker-"):
            return int(tag[7:])
    return 0


# Print out stdout/stderr if it exists, always
def print_attachments(stream, test, all_channels=False, show_binary_attachments=False):
    """Print out subunit attachments.

    Print out subunit attachments that contain content. This
    runs in 2 modes, one for successes where we print out just stdout
    and stderr, and an override that dumps all the attachments.
    """
    channels = ("stdout", "stderr")
    for name, detail in test["details"].items():
        # NOTE(sdague): the subunit names are a little crazy, and actually
        # are in the form pythonlogging:'' (with the colon and quotes)
        name = name.split(":")[0]
        if detail.content_type.type == "test":
            detail.content_type.type = "text"
        if all_channels or name in channels:
            title = "Captured %s:" % name
            stream.write("\n{}\n{}\n".format(title, ("~" * len(title))))

            # indent attachment lines 4 spaces to make them visually
            # offset
            if detail.content_type.type == "text":
                for line in detail.iter_text():
                    stream.write("    %s\n" % line)
            elif show_binary_attachments:  # binary
                for line in detail.iter_bytes():
                    stream.write("    %s\n" % line)


def find_test_run_time_diff(test_id, run_time):
    times_db_path = os.path.join(
        os.path.join(os.getcwd(), ".testrepository"), "times.dbm"
    )
    if os.path.isfile(times_db_path):
        try:
            test_times = dbm.open(times_db_path)
        except Exception:
            return False
        try:
            avg_runtime = float(test_times.get(str(test_id), False))
        except Exception:
            try:
                avg_runtime = float(test_times[str(test_id)])
            except Exception:
                avg_runtime = False

        if avg_runtime and avg_runtime > 0:
            run_time = float(run_time.rstrip("s"))
            perc_diff = ((run_time - avg_runtime) / avg_runtime) * 100
            return perc_diff
    return False


def show_outcome(
    stream,
    test,
    print_failures=False,
    failonly=False,
    enable_diff=False,
    threshold="0",
    abbreviate=False,
    enable_color=False,
    suppress_attachments=False,
    all_attachments=False,
    show_binary_attachments=True,
):
    global RESULTS
    status = test["status"]
    # TODO(sdague): ask lifeless why on this?
    if status == "exists":
        return

    worker = find_worker(test)
    name = cleanup_test_name(test["id"])
    duration = get_duration(test["timestamps"])

    if worker not in RESULTS:
        RESULTS[worker] = []
    RESULTS[worker].append(test)

    # don't count the end of the return code as a fail
    if name == "process-returncode":
        return

    for color in [colorizer.AnsiColorizer, colorizer.NullColorizer]:
        if not enable_color:
            color = colorizer.NullColorizer(stream)
            break
        if color.supported():
            color = color(stream)
            break

    if status == "fail" or status == "uxsuccess":
        FAILS.append(test)
        if abbreviate:
            color.write("F", "red")
        else:
            stream.write("{{{}}} {} [{}] ... ".format(worker, name, duration))
            color.write("FAILED", "red")
            stream.write("\n")
            if not print_failures:
                print_attachments(
                    stream,
                    test,
                    all_channels=True,
                    show_binary_attachments=show_binary_attachments,
                )
    elif not failonly:
        if status == "success" or status == "xfail":
            if abbreviate:
                color.write(".", "green")
            else:
                out_string = "{{{}}} {} [{}".format(worker, name, duration)
                perc_diff = find_test_run_time_diff(test["id"], duration)
                if enable_diff:
                    if perc_diff and abs(perc_diff) >= abs(float(threshold)):
                        if perc_diff > 0:
                            out_string = out_string + " +%.2f%%" % perc_diff
                        else:
                            out_string = out_string + " %.2f%%" % perc_diff
                stream.write(out_string + "] ... ")
                color.write("ok", "green")
                stream.write("\n")
                if not suppress_attachments:
                    print_attachments(
                        stream,
                        test,
                        all_channels=all_attachments,
                        show_binary_attachments=show_binary_attachments,
                    )
        elif status == "skip":
            if abbreviate:
                color.write("S", "blue")
            else:
                reason = test["details"].get("reason", "")
                if reason:
                    reason = ": " + reason.as_text()
                stream.write("{{{}}} {} ... ".format(worker, name))
                color.write("SKIPPED", "blue")
                stream.write("%s" % (reason))
                stream.write("\n")
        else:
            if abbreviate:
                stream.write("%s" % test["status"][0])
            else:
                stream.write(
                    "{{{}}} {} [{}] ... {}\n".format(
                        worker, name, duration, test["status"]
                    )
                )
                if not print_failures:
                    print_attachments(
                        stream,
                        test,
                        all_channels=True,
                        show_binary_attachments=show_binary_attachments,
                    )

    stream.flush()


def print_fails(stream):
    """Print summary failure report."""
    if not FAILS:
        return
    stream.write("\n==============================\n")
    stream.write("Failed %s tests - output below:" % len(FAILS))
    stream.write("\n==============================\n")
    for f in FAILS:
        stream.write("\n%s\n" % f["id"])
        stream.write("%s\n" % ("-" * len(f["id"])))
        print_attachments(stream, f, all_channels=True)
    stream.write("\n")


def count_tests(key, value):
    count = 0
    for k, v in RESULTS.items():
        for item in v:
            if key in item:
                if re.search(value, item[key]):
                    count += 1
    return count


def get_stuck_in_progress():
    key = "status"
    match = re.compile("^inprogress$")
    in_progress = []
    for k, v in RESULTS.items():
        for item in v:
            if key in item:
                if match.search(item[key]):
                    in_progress.append(item["id"])
    return in_progress


def run_time():
    runtime = 0.0
    for k, v in RESULTS.items():
        for test in v:
            test_dur = get_duration(test["timestamps"]).strip("s")
            # NOTE(toabctl): get_duration() can return an empty string
            # which leads to a ValueError when casting to float
            if test_dur:
                runtime += float(test_dur)
    return runtime


def worker_stats(worker):
    tests = RESULTS[worker]
    num_tests = len(tests)
    stop_time = tests[-1]["timestamps"][1]
    start_time = tests[0]["timestamps"][0]
    if not start_time or not stop_time:
        delta = "N/A"
    else:
        delta = stop_time - start_time
    return num_tests, str(delta)


def print_summary(stream, elapsed_time):
    stream.write("\n======\nTotals\n======\n")
    stream.write(
        "Ran: {} tests in {:.4f} sec.\n".format(
            count_tests("status", ".*"), total_seconds(elapsed_time)
        )
    )
    stream.write(" - Passed: %s\n" % count_tests("status", "^success$"))
    stream.write(" - Skipped: %s\n" % count_tests("status", "^skip$"))
    stream.write(" - Expected Fail: %s\n" % count_tests("status", "^xfail$"))
    stream.write(" - Unexpected Success: %s\n" % count_tests("status", "^uxsuccess$"))
    stream.write(" - Failed: %s\n" % count_tests("status", "^fail$"))
    stream.write("Sum of execute time for each test: %.4f sec.\n" % run_time())

    # we could have no results, especially as we filter out the process-codes
    if RESULTS:
        stream.write("\n==============\nWorker Balance\n==============\n")

        for w in range(max(RESULTS.keys()) + 1):
            if w not in RESULTS:
                stream.write(" - WARNING: missing Worker %s!\n" % w)
            else:
                num, time = worker_stats(w)
                out_str = " - Worker {} ({} tests) => {}".format(w, num, time)
                if time.isdigit():
                    out_str += "s"
                out_str += "\n"
                stream.write(out_str)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version="%s" % __version__)
    parser.add_argument(
        "--no-failure-debug",
        "-n",
        action="store_true",
        dest="print_failures",
        help="Disable printing failure " "debug information in realtime",
    )
    parser.add_argument(
        "--fails",
        "-f",
        action="store_true",
        dest="post_fails",
        help="Print failure debug " "information after the stream is proccesed",
    )
    parser.add_argument(
        "--failonly",
        action="store_true",
        dest="failonly",
        help="Don't print success items",
        default=(os.environ.get("TRACE_FAILONLY", False) is not False),
    )
    parser.add_argument(
        "--abbreviate",
        "-a",
        action="store_true",
        dest="abbreviate",
        help="Print one character status" "for each test",
    )
    parser.add_argument(
        "--perc-diff",
        "-d",
        action="store_true",
        dest="enable_diff",
        help="Print percent change in run time on each test ",
    )
    parser.add_argument(
        "--diff-threshold",
        "-t",
        dest="threshold",
        help="Threshold to use for displaying percent change "
        "from the avg run time. If one is not specified "
        "the percent change will always be displayed",
    )
    parser.add_argument(
        "--no-summary",
        action="store_true",
        help="Don't print the summary of the test run after " " completes",
    )
    parser.add_argument(
        "--color", action="store_true", help="Print results with colors"
    )
    return parser.parse_args()


def trace(
    stdin,
    stdout,
    print_failures=False,
    failonly=False,
    enable_diff=False,
    abbreviate=False,
    color=False,
    post_fails=False,
    no_summary=False,
    suppress_attachments=False,
    all_attachments=False,
    show_binary_attachments=False,
):
    stream = subunit.ByteStreamToStreamResult(stdin, non_subunit_name="stdout")
    outcomes = testtools.StreamToDict(
        functools.partial(
            show_outcome,
            stdout,
            print_failures=print_failures,
            failonly=failonly,
            enable_diff=enable_diff,
            abbreviate=abbreviate,
            enable_color=color,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=show_binary_attachments,
        )
    )
    summary = testtools.StreamSummary()
    result = testtools.CopyStreamResult([outcomes, summary])
    result = testtools.StreamResultRouter(result)
    cat = subunit.test_results.CatFiles(stdout)
    result.add_rule(cat, "test_id", test_id=None)
    result.startTestRun()
    try:
        stream.run(result)
    finally:
        result.stopTestRun()
    start_times = []
    stop_times = []
    for worker in RESULTS:
        start_times += [
            x["timestamps"][0]
            for x in RESULTS[worker]
            if x["timestamps"][0] is not None
        ]
        stop_times += [
            x["timestamps"][1]
            for x in RESULTS[worker]
            if x["timestamps"][1] is not None
        ]
    if print_full_output(stdout, start_times, stop_times, post_fails, no_summary):
        return 1
    return 0 if results.wasSuccessful(summary) else 1


def print_full_output(stdout, start_times, stop_times, post_fails, no_summary):
    """Print output plus edge case validation"""
    if not start_times or count_tests("status", ".*") == 0:
        print("The test run didn't actually run any tests", file=sys.stderr)
        return 1

    start_time = min(start_times)
    stop_time = max(stop_times)
    elapsed_time = stop_time - start_time

    if post_fails:
        print_fails(stdout)
    if not no_summary:
        print_summary(stdout, elapsed_time)

    # NOTE(mtreinish): Ideally this should live in testtools streamSummary
    # this is just in place until the behavior lands there (if it ever does)
    if count_tests("status", "^xfail$") + count_tests("status", "^success$") == 0:
        print("\nNo tests were successful during the run", file=sys.stderr)
        return 1
    in_progress = get_stuck_in_progress()
    if in_progress:
        print(
            "\nThe following tests exited without returning a status\n"
            "and likely segfaulted or crashed Python:",
            file=sys.stderr,
        )
        for test in in_progress:
            print("\n\t* %s" % test, file=sys.stderr)
        return 1


def main():
    args = parse_args()
    exit(
        trace(
            sys.stdin,
            sys.stdout,
            args.print_failures,
            args.failonly,
            args.enable_diff,
            args.abbreviate,
            args.color,
            args.post_fails,
            args.no_summary,
        )
    )


if __name__ == "__main__":
    main()
